from typing import Union, Any, Callable

from beartype import beartype
from beartype.roar import BeartypeCallHintParamViolation
from typeguard import check_type
from typing_inspect import get_origin, get_args

from ofrak.service.serialization.pjson_types import PJSONType
from ofrak.service.serialization.serializers.serializer_i import SerializerInterface


class UnionSerializer(SerializerInterface):
    """
    Serialize and deserialize `Union[...]` into `PJSONType`.

    This includes Optional, as Optional[X] == Union[X, type(None)]

    Implementation: all types in the Union are tried in order, and the first for which
    handling doesn't return an error is used.
    """

    targets = (lambda type_hint: get_origin(type_hint) == Union,)

    def to_pjson_checking_type(self, obj: Any, type_hint: Any) -> PJSONType:
        check_type("obj", obj, type_hint)
        return self._service.to_pjson(obj, type_hint)

    def from_pjson_checking_type(self, pjson_obj: PJSONType, type_hint: Any) -> Any:
        obj = self._service.from_pjson(pjson_obj, type_hint)
        check_type("obj", obj, type_hint)
        return obj

    @beartype
    def obj_to_pjson(self, obj: Any, type_hint: Any) -> PJSONType:
        return self._try_all_types(obj, type_hint, self.to_pjson_checking_type)

    @beartype
    def pjson_to_obj(self, pjson_obj: PJSONType, type_hint: Any) -> Any:
        return self._try_all_types(pjson_obj, type_hint, self.from_pjson_checking_type)

    def _try_all_types(self, obj: Any, type_hint: Any, handler: Callable[[Any, Any], Any]) -> Any:
        args = get_args(type_hint)
        failure_reasons = dict()
        for arg in args:
            try:
                return handler(obj, arg)
            except (TypeError, BeartypeCallHintParamViolation, AttributeError, KeyError) as e:
                failure_reasons[arg] = e
        reasons_string = "\n".join(f"{arg}: {reason}" for arg, reason in failure_reasons.items())
        raise TypeError(
            f"Couldn't run {handler.__name__} on {obj} with any of the types:\n{reasons_string}"
        )



import logging
import struct
from enum import IntEnum
from dataclasses import dataclass
from typing import List, Tuple

from ofrak import Identifier
from ofrak.core.code_region import CodeRegion

from ofrak.resource import Resource
from ofrak.model.resource_model import ResourceAttributes
from ofrak.component.unpacker import Unpacker
from ofrak.component.packer import Packer
from ofrak.core.binary import GenericBinary
from ofrak_type.range import Range
from ofrak.service.resource_service_i import ResourceFilter

LOGGER = logging.getLogger(__name__)

UF2_MAGIC_START_ONE = 0x0A324655
UF2_MAGIC_START_TWO = 0x9E5D5157
UF2_MAGIC_END = 0x0AB16F30

HEADER_LENGTH = 32
DATA_LENGTH = 476


@dataclass
class Uf2File(GenericBinary):
    """
    A UF2 file
    """


@dataclass(**ResourceAttributes.DATACLASS_PARAMS)
class Uf2FileAttributes(ResourceAttributes):
    """
    Remembers all the information needed to repack that can't be deduced from the contents.
    """

    family_id: int


@dataclass(**ResourceAttributes.DATACLASS_PARAMS)
class Uf2BlockHeader(ResourceAttributes):
    """
    Recreates the official spec

    Offset  Size    Value
    0       4       First magic number, 0x0A324655 ("UF2\n")
    4       4       Second magic number, 0x9E5D5157
    8       4       Flags
    12      4       Address in flash where the data should be written
    16      4       Number of bytes used in data (often 256)
    20      4       Sequential block number; starts at 0
    24      4       Total number of blocks in file
    28      4       File size or board family ID or zero
    32      476     Data, padded with zeros
    508     4       Final magic number, 0x0AB16F30
    """

    flags: int
    target_addr: int
    payload_size: int
    block_no: int
    num_blocks: int
    filesize_family_id: int


class Uf2Flags(IntEnum):
    NOT_MAIN_FLASH = 0x1
    FILE_CONTAINER = 0x1000
    FAMILY_ID_PRESENT = 0x2000
    MD5_CHECKSUM_PRESENT = 0x4000
    EXTENSION_TAGS_PRESENT = 0x8000


class Uf2Unpacker(Unpacker[None]):
    """
    UF2 unpacker.

    Extracts the data from a UF2 packed file.
    """

    targets = (Uf2File,)
    children = (CodeRegion,)

    async def unpack(self, resource: Resource, config=None):
        """
        Unpack a UF2 file.

        UF2 files contain blocks of binary data.
        """
        data_length = await resource.get_data_length()

        ranges: List[Tuple[Range, bytes]] = []

        # block_no are 0 indexed, to make the check do one fewer check, we start at -1
        previous_block_no = -1
        family_id = None
        file_num_blocks = None
        block_no = 0

        for i in range(0, data_length, 512):
            data = await resource.get_data(Range(i, (i + 512)))
            (
                magic_start_one,
                magic_start_two,
                flags,
                target_addr,
                payload_size,
                block_no,
                num_blocks,
                filesize_familyID,
                payload_data,
                magic_end,
            ) = struct.unpack("8I476sI", data)

            # basic sanity checks
            if magic_start_one != UF2_MAGIC_START_ONE:
                raise ValueError("Bad Start Magic")
            if magic_start_two != UF2_MAGIC_START_TWO:
                raise ValueError("Bad Start Magic")
            if magic_end != UF2_MAGIC_END:
                raise ValueError("Bad End Magic")

            if (previous_block_no - block_no) != -1:
                raise ValueError("Skipped a block number")
            previous_block_no = block_no

            if not file_num_blocks:
                file_num_blocks = num_blocks

            if family_id is None:
                family_id = filesize_familyID
            else:
                if family_id != filesize_familyID:
                    raise NotImplementedError("Multiple family IDs in file not supported")

            # unpack data
            if flags & Uf2Flags.NOT_MAIN_FLASH:
                # data not written to main flash
                raise NotImplementedError(
                    "Data not written to main flash is currently not supported"
                )
            elif flags & Uf2Flags.FILE_CONTAINER:
                # file container
                raise NotImplementedError("File containers are currently not implemented")
            elif flags & Uf2Flags.FAMILY_ID_PRESENT:
                data = payload_data[0:payload_size]
                if len(ranges) == 0:
                    ranges.append((Range(target_addr, target_addr + payload_size), data))
                else:
                    last_region_range, last_region_data = ranges[-1]

                    # if range is adjacent, extend, otherwise start a new one
                    if target_addr - last_region_range.end == 0:
                        last_region_range.end = target_addr + payload_size
                        last_region_data += data
                        ranges[-1] = (last_region_range, last_region_data)
                    else:
                        ranges.append((Range(target_addr, target_addr + payload_size), data))
            else:
                # unsupported flags
                raise ValueError(f"Unsupported flags {flags}")

        # count vs 0 indexed (there are 256 blocks from 0-255)
        if file_num_blocks != (block_no + 1):
            raise ValueError(
                f"Incorrect number of blocks. Expected {file_num_blocks}, got {block_no}"
            )

        if family_id:
            file_attributes = Uf2FileAttributes(family_id)
            resource.add_attributes(file_attributes)

        # print("num: ", last_block_no, "file: ", file_num_blocks)
        # assert last_block_no == file_num_blocks, "Did not unpack enough blocks"

        for flash_range, flash_data in ranges:
            await resource.create_child_from_view(
                CodeRegion(flash_range.start, flash_range.end - flash_range.start),
                data=flash_data,
            )


class Uf2FilePacker(Packer[None]):
    """
    Pack a resource into the UF2 file format
    """

    id = b"Uf2FilePacker"
    targets = (Uf2File,)

    async def pack(self, resource: Resource, config=None):
        """
        Pack a resource into a UF2 file

        :param resource:
        :param config:
        """

        payloads: List[Tuple[int, int, bytes]] = []  # List of target_addr, payload_data

        for memory_region_r in await resource.get_children(
            r_filter=ResourceFilter(
                tags=(CodeRegion,),
            )
        ):
            memory_region = await memory_region_r.view_as(CodeRegion)
            data = await memory_region_r.get_data()
            data_length = await memory_region_r.get_data_length()
            data_range = memory_region.vaddr_range()
            addr = data_range.start

            for i in range(0, data_length, 256):
                payloads.append((addr + i, 256, data[i : (i + 256)]))
                continue

        num_blocks = len(payloads)
        block_no = 0

        file_attributes = resource.get_attributes(attributes_type=Uf2FileAttributes)
        family_id = file_attributes.family_id

        repacked_data = b""

        for target_addr, payload_size, payload_data in payloads:
            repacked_data += struct.pack(
                "8I476sI",
                UF2_MAGIC_START_ONE,
                UF2_MAGIC_START_TWO,
                Uf2Flags.FAMILY_ID_PRESENT,
                target_addr,
                payload_size,
                block_no,
                num_blocks,
                family_id,
                payload_data + b"\x00" * (467 - payload_size),  # add padding
                UF2_MAGIC_END,
            )
            block_no += 1

        resource.queue_patch(Range(0, await resource.get_data_length()), repacked_data)


class Uf2FileIdentifier(Identifier[None]):
    targets = (GenericBinary,)

    async def identify(self, resource: Resource, config=None) -> None:
        if await resource.get_data_length() < 8:
            pass
        else:
            data = await resource.get_data(Range(0, 8))
            magic_one, magic_two = struct.unpack("<II", data)
            if magic_one == UF2_MAGIC_START_ONE and magic_two == UF2_MAGIC_START_TWO:
                resource.add_tag(Uf2File)



# Suppress all BeartypeDecorHintPep585DeprecationWarning warnings pending resolution of
# https://github.com/redballoonsecurity/ofrak/issues/92.
# This is nice for OFRAK users, who don't want to see these warnings.
from beartype.roar import BeartypeDecorHintPep585DeprecationWarning
from warnings import filterwarnings

filterwarnings("ignore", category=BeartypeDecorHintPep585DeprecationWarning)



import functools
import logging
from typing import Set, List, Iterable, Dict, cast

from ofrak.model.component_model import ComponentContext
from ofrak.model.data_model import DataPatchesResult, DataModel
from ofrak.model.resource_model import (
    ResourceContext,
    ResourceAttributeDependency,
    MutableResourceModel,
    Data,
)
from ofrak.service.data_service_i import DataServiceInterface
from ofrak.service.resource_service_i import ResourceServiceInterface
from ofrak_type.range import Range

LOGGER = logging.getLogger(__file__)


class DependencyHandler:
    """
    Stateless handler for dealing with creating and invalidating dependencies. Intended for
    one-time use in a component, or possible re-use in a resource.
    """

    def __init__(
        self,
        resource_service: ResourceServiceInterface,
        data_service: DataServiceInterface,
        component_context: ComponentContext,
        resource_context: ResourceContext,
    ):
        self._resource_service = resource_service
        self._data_service = data_service
        self._component_context = component_context
        self._resource_context = resource_context

    @functools.lru_cache(None)
    async def map_data_ids_to_resources(
        self, data_ids: Iterable[bytes]
    ) -> Dict[bytes, MutableResourceModel]:
        resources_by_data_id = dict()
        for resource_id, resource_m in self._resource_context.resource_models.items():
            if resource_m.data_id is not None:
                resources_by_data_id[resource_m.data_id] = resource_m

        missing_data_ids = set()
        for data_id in data_ids:
            if data_id not in resources_by_data_id:
                missing_data_ids.add(data_id)

        missing_resources = await self._resource_service.get_by_data_ids(missing_data_ids)
        for missing_resource in missing_resources:
            missing_resource_data_id = cast(bytes, missing_resource.data_id)
            if missing_resource_data_id in resources_by_data_id:
                raise ValueError("Something is wrong in the implementation")
            mutable_resource = MutableResourceModel.from_model(missing_resource)
            resources_by_data_id[missing_resource_data_id] = mutable_resource
            self._resource_context.resource_models[missing_resource.id] = mutable_resource

        return resources_by_data_id

    async def handle_post_patch_dependencies(
        self, patch_results: List[DataPatchesResult]
    ) -> Set[MutableResourceModel]:
        modified_resources = set()

        # Create look up maps for resources and dependencies
        resources_by_data_id = await self.map_data_ids_to_resources(
            patch_result.data_id for patch_result in patch_results
        )

        all_data_ids = [patch_result.data_id for patch_result in patch_results]
        models_by_data_id: Dict[bytes, DataModel] = {
            data_id: model
            for data_id, model in zip(
                all_data_ids, await self._data_service.get_by_ids(all_data_ids)
            )
        }
        # Go through and update all models' Data
        for data_patch_result in patch_results:
            resource_m = resources_by_data_id[data_patch_result.data_id]
            data_m = models_by_data_id[data_patch_result.data_id]
            resource_m.add_attributes(Data(data_m.range.start, data_m.range.length()))
            modified_resources.add(resource_m)

        unhandled_dependencies: Set[ResourceAttributeDependency] = set()
        # Figure out which components results must be invalidated based on data changes

        for data_patch_result in patch_results:
            resource_m = resources_by_data_id[data_patch_result.data_id]
            removed_data_dependencies = set()
            # Iterate over the resource's data dependencies to find one that's affected by one of
            # the patch range
            for dependency, dependency_ranges in resource_m.data_dependencies.items():
                # Iterate over the resource's data dependency ranges to find a range that overlaps
                # the patch range
                for dependency_range in dependency_ranges:
                    for patch_range in data_patch_result.patches:
                        if not dependency_range.overlaps(patch_range):
                            continue
                        LOGGER.debug(
                            f"Invalidating results of {dependency.component_id!r} on resource "
                            f"{dependency.dependent_resource_id.hex()} due to a data change on "
                            f"resource {resource_m.id.hex()}"
                        )
                        # That dependency is invalid, we can move on to the next dependency
                        unhandled_dependencies.add(dependency)
                        removed_data_dependencies.add(dependency)
                        break
                    # If a previous dependency range was found to affect the dependency, no need to
                    # continue iterating the ranges
                    if dependency in removed_data_dependencies:
                        break
            for removed_data_dependency in removed_data_dependencies:
                resource_m.remove_dependency(removed_data_dependency)
                modified_resources.add(resource_m)

        # Recursively invalidate component results based on other components that were invalidated
        handled_dependencies: Set[ResourceAttributeDependency] = set()

        await self._invalidate_dependencies(
            handled_dependencies, unhandled_dependencies, modified_resources
        )

        return modified_resources

    def create_component_dependencies(
        self,
        component_id: bytes,
        component_version: int,
    ):
        """
        Register dependencies between the component and the resources it interacts with.

        This may not even be necessary since Resource.add_attributes does this anyway...
        """
        self._validate_resource_context_complete()

        for resource_id in self._component_context.resources_created:
            new_resource_m = self._resource_context.resource_models[resource_id]
            for attributes in new_resource_m.attributes.keys():
                new_resource_m.add_component_for_attributes(
                    component_id, component_version, attributes
                )

    def create_resource_dependencies(
        self,
        component_id: bytes,
    ):
        """
        Register dependencies between a resource with some attributes and the resources which
        were accessed in the context where these attributes were added.

        When a component runs, this method is called to record what data was accessed by the
        component and what resource attributes from other resources were accessed within that
        component. These registered dependencies allow for OFRAK to not rerun analyzers when the
        resource and its dependencies have not changed.

        Whenever a [Modifier][ofrak.component.modifier.Modifier] is run, these resource attribute
        dependencies are invalidated so as to force analysis to be rerun.

        :param bytes component_id:
        """
        self._validate_resource_context_complete()
        resource_dependencies = []

        # Create dependency for each attribute on newly created resources
        for resource_id in self._component_context.resources_created:
            new_resource_m = self._resource_context.resource_models[resource_id]
            for attributes in new_resource_m.attributes.keys():
                resource_dependencies.append(
                    ResourceAttributeDependency(
                        resource_id,
                        component_id,
                        attributes,
                    )
                )

        # Create dependency for each new attribute on modified resources
        for resource_id in self._component_context.modification_trackers.keys():
            modified_resource_m = self._resource_context.resource_models[resource_id]
            for attrs_added in modified_resource_m.diff.attributes_added.keys():
                resource_dependencies.append(
                    ResourceAttributeDependency(
                        resource_id,
                        component_id,
                        attrs_added,
                    )
                )

        # Add dependencies to all resources which were accessed
        for resource_id, access_tracker in self._component_context.access_trackers.items():
            if resource_id in self._component_context.resources_created:
                # Avoid all the newly created components depending on each other
                continue
            accessed_resource_m = self._resource_context.resource_models[resource_id]
            merged_accessed_ranges = Range.merge_ranges(access_tracker.data_accessed)
            # Add attributes dependency on all accessed attributes
            for attributes_accessed in access_tracker.attributes_accessed:
                for resource_dependency in resource_dependencies:
                    accessed_resource_m.add_attribute_dependency(
                        attributes_accessed, resource_dependency
                    )
            # Add data dependency on all accessed data
            for accessed_range in merged_accessed_ranges:
                for resource_dependency in resource_dependencies:
                    accessed_resource_m.add_data_dependency(resource_dependency, accessed_range)

    def _validate_resource_context_complete(self):
        for resource_id in self._component_context.resources_created:
            if resource_id not in self._resource_context.resource_models:
                raise ValueError(
                    f"The resource model {resource_id.hex()} was created but it's not in the "
                    f"resource context"
                )

    async def _fetch_missing_resources(self, resource_ids: Iterable[bytes]):
        missing_resource_ids = set()
        # Fetch all the resources referred to by the unhandled dependencies
        for resource_id in resource_ids:
            if resource_id not in self._resource_context.resource_models:
                missing_resource_ids.add(resource_id)
        missing_resources = await self._resource_service.get_by_ids(missing_resource_ids)
        for missing_resource in missing_resources:
            self._resource_context.resource_models[
                missing_resource.id
            ] = MutableResourceModel.from_model(missing_resource)

    async def _invalidate_dependencies(
        self,
        handled_dependencies: Set[ResourceAttributeDependency],
        unhandled_dependencies: Set[ResourceAttributeDependency],
        resources_modified: Set[MutableResourceModel],
    ):
        """
        Invalidate the unhandled resource attribute dependencies.

        :param Set[ResourceAttributeDependency] handled_dependencies: A set of dependencies that
        have already been invalidated
        :param Set[ResourceAttributeDependency] unhandled_dependencies: A set of dependencies
        that should be invalidated
        """
        if len(unhandled_dependencies) == 0:
            return

        dependent_resource_ids = {
            dependency.dependent_resource_id for dependency in unhandled_dependencies
        }

        deleted_dependent_ids = {
            r_id
            for r_id, currently_exists in zip(
                dependent_resource_ids,
                await self._resource_service.verify_ids_exist(dependent_resource_ids),
            )
            if not currently_exists
        }
        await self._fetch_missing_resources(
            dependent_resource_ids.difference(deleted_dependent_ids)
        )

        # Invalidate the resources' attributes referred to by the unhandled_dependencies
        next_unhandled_dependencies = set()
        for dependency in unhandled_dependencies:
            # It's possible that the attribute was already invalidated from an earlier run
            if dependency in handled_dependencies:
                continue

            # If the dependent resource was deleted, don't need to propagate dependency invalidation
            if dependency.dependent_resource_id in deleted_dependent_ids:
                handled_dependencies.add(dependency)
                continue

            try:
                resource_m = self._resource_context.resource_models[
                    dependency.dependent_resource_id
                ]
            except KeyError as e:
                missing_model = await self._resource_service.get_by_id(
                    dependency.dependent_resource_id
                )
                resource_m = MutableResourceModel.from_model(missing_model)

            # Invalidate the attributes on the resource
            handled_dependencies.add(dependency)

            # The component id is not necessarily present. It could have been invalidated already
            # by a previous patch that impacted other resources that this resource depends on.
            if resource_m.get_component_id_by_attributes(dependency.attributes):
                resource_m.remove_component(dependency.component_id, dependency.attributes)
                self._component_context.mark_resource_modified(resource_m.id)
                resources_modified.add(resource_m)

            # Find other dependencies to invalidate due to the invalidation of the attributes
            invalidated_dependencies = set()
            for next_dependency in resource_m.attribute_dependencies[dependency.attributes]:
                # Make sure the dependency wasn't already handled
                if next_dependency not in handled_dependencies:
                    LOGGER.debug(
                        f"Invalidating attributes {next_dependency.attributes.__name__} from "
                        f"component {next_dependency.component_id!r} on resource "
                        f"{next_dependency.dependent_resource_id.hex()} due to "
                        f"attributes {dependency.attributes.__name__} on resource"
                        f" {dependency.dependent_resource_id.hex()} being invalidated"
                    )

                    invalidated_dependencies.add(next_dependency)

            for invalidated_dependency in invalidated_dependencies:
                resource_m.remove_dependency(invalidated_dependency)
                self._component_context.mark_resource_modified(resource_m.id)
                resources_modified.add(resource_m)
            next_unhandled_dependencies.update(invalidated_dependencies)

        await self._invalidate_dependencies(
            handled_dependencies,
            next_unhandled_dependencies,
            resources_modified,
        )


class DependencyHandlerFactory:
    def create(
        self,
        resource_service: ResourceServiceInterface,
        data_service: DataServiceInterface,
        component_context: ComponentContext,
        resource_context: ResourceContext,
    ) -> DependencyHandler:
        return DependencyHandler(
            resource_service, data_service, component_context, resource_context
        )



import asyncio
import re
from dataclasses import dataclass

from ofrak.component.analyzer import Analyzer
from ofrak.component.modifier import Modifier, ModifierError
from ofrak.component.unpacker import Unpacker
from ofrak.core.binary import BinaryPatchConfig, BinaryPatchModifier, GenericText, GenericBinary
from ofrak.core.code_region import CodeRegion
from ofrak.core.filesystem import File
from ofrak.core.program_section import ProgramSection
from ofrak.model.component_model import ComponentConfig
from ofrak.model.resource_model import index
from ofrak.model.viewable_tag_model import AttributesType
from ofrak.resource import Resource
from ofrak.resource_view import ResourceView
from ofrak_type import Range


@dataclass
class StringPatchingConfig(ComponentConfig):
    """
    Dataclass required to apply a string patch with `StringPatchingModifier`. The configuration
    describes the `offset` where the patch is to be applied, and the `string` to patch in.

    :var offset: the offset at which to apply the patch
    :var string: the string to patch in
    """

    offset: int
    string: str
    null_terminate: bool = False


class StringPatchingModifier(Modifier[StringPatchingConfig]):
    """
    Patch a string in a resource at a given offset, based on the provided configuration.
    """

    id = b"StringPatchingModifier"
    targets = (GenericText,)

    async def modify(self, resource: Resource, config: StringPatchingConfig):
        new_data = config.string.encode("utf-8")
        if config.null_terminate:
            new_data += b"\x00"
        patch_config = BinaryPatchConfig(config.offset, new_data)
        await resource.run(BinaryPatchModifier, patch_config)


@dataclass
class StringFindReplaceConfig(ComponentConfig):
    """
    :var to_find: the string to search for
    :var replace_with: the string to pass in
    :var null_terminate: add a null terminator to the replacement if True
    :var allow_overflow: allow the replace string to overflow the found string if True
    """

    to_find: str
    replace_with: str
    null_terminate: bool = True
    allow_overflow: bool = False


class StringFindReplaceModifier(Modifier[StringFindReplaceConfig]):
    """
    Find and replace all instances of a given string with a replacement string.
    """

    targets = (GenericBinary, File)

    async def modify(self, resource: Resource, config: StringFindReplaceConfig) -> None:
        to_find = config.to_find.encode("utf-8")
        replace_with = config.replace_with.encode("utf-8") + (
            b"\x00" if config.null_terminate and config.replace_with[-1] != "\x00" else b""
        )
        if not config.allow_overflow and len(replace_with) > len(to_find):
            raise ModifierError(
                f"Original string is longer than the new string ({len(to_find)} < "
                f"{len(replace_with)})! Set config.allow_overflow = True to override this error. "
                f"If you expect that the string to replace is null-terminated, then an overflow "
                f"of one byte when config.null_terminate = True will not have any effect."
            )
        for offset in await resource.search_data(to_find):
            await resource.run(BinaryPatchModifier, BinaryPatchConfig(offset, replace_with))


@dataclass
class AsciiString(ResourceView):
    """
    Resource representing a C-style, NULL-terminated string of ASCII characters. The `text` string
    is not NULL-terminated.
    """

    text: str

    @index
    def Text(self) -> str:
        return self.text

    @classmethod
    def caption(cls, all_attributes) -> str:
        try:
            s = all_attributes[AttributesType[AsciiString]]
            return f"string: '{s.text}'"
        except KeyError:
            return super().caption(all_attributes)


class AsciiStringAnalyzer(Analyzer[None, AsciiString]):
    """
    Extract the decoded string from a C-style, NULL-terminated string of ASCII characters.
    """

    targets = (AsciiString,)
    outputs = (AsciiString,)

    async def analyze(self, resource: Resource, config: None) -> AsciiString:
        raw_without_null_byte = (await resource.get_data()).rstrip(b"\x00")
        return AsciiString(raw_without_null_byte.decode("ascii"))


class StringsUnpacker(Unpacker[None]):
    """
    Unpack NULL-terminated strings of printable ASCII characters.
    """

    targets = (ProgramSection,)  # TODO: Other reasonable targets?
    children = (AsciiString,)

    # match sequences of at least 2 (or 8 in CodeRegions) printable characters ending with NULL
    # printable characters defined as: ASCII between ' ' and '~', tab, newline, carriage return
    LONG_STRING_PATTERN = re.compile(b"([ -~\n\t\r]{8,})\x00")
    SHORT_STRING_PATTERN = re.compile(b"([ -~\n\t\r]{2,})\x00")

    async def unpack(self, resource: Resource, config: None) -> None:
        if resource.get_data_id() is None:
            return
        if resource.has_tag(CodeRegion):
            # code is less likely to have strings so more likely to have false positives
            pattern = self.LONG_STRING_PATTERN
        else:
            pattern = self.SHORT_STRING_PATTERN

        children = [
            resource.create_child_from_view(
                AsciiString(string.rstrip(b"\x00").decode("ascii")),
                data_range=Range.from_size(offset, len(string)),
            )
            for offset, string in await resource.search_data(pattern)
        ]

        await asyncio.gather(*children)



import asyncio
import tempfile
from concurrent.futures.process import ProcessPoolExecutor
from dataclasses import dataclass
from typing import Dict

from ofrak.resource import ResourceFactory, Resource

from ofrak.model.resource_model import ResourceAttributes

from ofrak.component.abstract import ComponentMissingDependencyError
from ofrak.component.analyzer import Analyzer

try:
    import binwalk

    BINWALK_INSTALLED = True
except ImportError:
    BINWALK_INSTALLED = False

from ofrak.core.binary import GenericBinary
from ofrak.core.filesystem import File
from ofrak.model.component_model import ComponentExternalTool
from ofrak.service.data_service_i import DataServiceInterface
from ofrak.service.resource_service_i import ResourceServiceInterface


class _BinwalkExternalTool(ComponentExternalTool):
    def __init__(self):
        super().__init__(
            "binwalk",
            "https://github.com/ReFirmLabs/binwalk",
            install_check_arg="",
        )

    async def is_tool_installed(self) -> bool:
        return BINWALK_INSTALLED


BINWALK_TOOL = _BinwalkExternalTool()


@dataclass(**ResourceAttributes.DATACLASS_PARAMS)
class BinwalkAttributes(ResourceAttributes):
    offsets: Dict[int, str]


class BinwalkAnalyzer(Analyzer[None, BinwalkAttributes]):
    targets = (GenericBinary, File)
    outputs = (BinwalkAttributes,)
    external_dependencies = (BINWALK_TOOL,)

    def __init__(
        self,
        resource_factory: ResourceFactory,
        data_service: DataServiceInterface,
        resource_service: ResourceServiceInterface,
    ):
        super().__init__(resource_factory, data_service, resource_service)
        self.pool = ProcessPoolExecutor()

    async def analyze(self, resource: Resource, config=None) -> BinwalkAttributes:
        if not BINWALK_INSTALLED:
            raise ComponentMissingDependencyError(self, BINWALK_TOOL)
        with tempfile.NamedTemporaryFile() as temp_file:
            data = await resource.get_data()
            temp_file.write(data)
            temp_file.flush()

            # Should errors be handled the way they are in the `DataSummaryAnalyzer`? Likely to be
            # overkill here.
            offsets = await asyncio.get_running_loop().run_in_executor(
                self.pool, _run_binwalk_on_file, temp_file.name
            )
        return BinwalkAttributes(offsets)


def _run_binwalk_on_file(filename):  # pragma: no cover
    offsets = dict()
    for module in binwalk.scan(filename, signature=True):
        for result in module.results:
            offsets[result.offset] = result.description
    return offsets



import logging
import re
import sys
from dataclasses import dataclass
from typing import List, Union, Tuple, Any

from ofrak.component.abstract import ComponentMissingDependencyError

try:
    from bincopy import BinFile

    BINCOPY_INSTALLED = True
except ImportError:
    BINCOPY_INSTALLED = False

from ofrak.component.analyzer import Analyzer
from ofrak.component.identifier import Identifier
from ofrak.component.packer import Packer
from ofrak.component.unpacker import Unpacker
from ofrak.core.binary import GenericBinary, GenericText
from ofrak.core.program_section import ProgramSection
from ofrak.core.program import Program
from ofrak.model.component_model import ComponentExternalTool
from ofrak.resource import Resource
from ofrak.service.resource_service_i import ResourceFilter
from ofrak_type.range import Range

LOGGER = logging.getLogger(__name__)


@dataclass
class Ihex(GenericBinary):
    """
    Intel HEX is a binary blob packaging format encoded in ASCII. It splits binary data into records,
    which are lines of ASCII representing in hex the byte count, address, type, checksums of stored data.
    It is typically used for flashing firmware.

    # printf "Hello world!" | bin2hex.py -
    :0C00000048656C6C6F20776F726C642197
    :00000001FF
    """


@dataclass
class IhexProgram(Program):
    address_limits: Range
    start_addr: Union[None, int]
    segments: List[Range]


_BINCOPY_TOOL = ComponentExternalTool(
    "bincopy",
    "https://github.com/eerimoq/bincopy",
    "--help",
)


class IhexAnalyzer(Analyzer[None, IhexProgram]):
    """
    Extract Intel HEX parameters
    """

    targets = (IhexProgram,)
    outputs = (IhexProgram,)

    external_dependencies = (_BINCOPY_TOOL,)

    async def analyze(self, resource: Resource, config: None = None) -> IhexProgram:
        raw_ihex = await resource.get_parent()
        ihex_program, _ = _binfile_analysis(await raw_ihex.get_data(), self)
        return ihex_program


class IhexUnpacker(Unpacker[None]):
    """
    Unpack an Intel HEX file, converting into raw bytes with padding bytes added to fill the gaps
    between segments. The result is a Program made up of a binary blob representing the entire
    memory space that the ihex file would load.
    """

    targets = (Ihex,)
    children = (IhexProgram,)

    external_dependencies = (_BINCOPY_TOOL,)

    async def unpack(self, resource: Resource, config=None):
        ihex_program, binfile = _binfile_analysis(await resource.get_data(), self)

        await resource.create_child_from_view(ihex_program, data=bytes(binfile.as_binary()))


class IhexProgramUnpacker(Unpacker[None]):
    """
    Unpack the individual segments from an Intel HEX Program's binary blob.
    """

    targets = (IhexProgram,)
    children = (ProgramSection,)

    async def unpack(self, resource: Resource, config=None):
        ihex_program = await resource.view_as(IhexProgram)
        for seg_vaddr_range in ihex_program.segments:
            # Segment is mapped into the program at an offset starting at the difference between
            # the segment's vaddr range and the program's base address
            segment_data_range = seg_vaddr_range.translate(-ihex_program.address_limits.start)
            await resource.create_child_from_view(
                ProgramSection(seg_vaddr_range.start, seg_vaddr_range.length()),
                data_range=segment_data_range,
            )


class IhexProgramPacker(Packer[None]):
    """
    Pack the segments of an Intel HEX program back into a binary blob. Recomputes segment size and
    program address range based on the actual segments at the time of packing.
    """

    targets = (IhexProgram,)

    async def pack(self, resource: Resource, config=None) -> None:
        updated_segments = []
        min_vaddr = sys.maxsize
        max_vaddr = 0
        for segment_r in await resource.get_children_as_view(
            ProgramSection, r_filter=ResourceFilter.with_tags(ProgramSection)
        ):
            seg_length = await segment_r.resource.get_data_length()
            seg_start = segment_r.virtual_address
            updated_segments.append(Range.from_size(seg_start, seg_length))
            min_vaddr = min(min_vaddr, seg_start)
            max_vaddr = max(max_vaddr, seg_start + seg_length)
        ihex_prog = await resource.view_as(IhexProgram)
        ihex_prog.segments = updated_segments
        ihex_prog.address_limits = Range(min_vaddr, max_vaddr)
        resource.add_view(ihex_prog)


class IhexPacker(Packer[None]):
    """
    Pack a binary blob representation of an Intel HEX program back into an Intel HEX file.
    """

    targets = (Ihex,)

    external_dependencies = (_BINCOPY_TOOL,)

    async def pack(self, resource: Resource, config=None) -> None:
        if not BINCOPY_INSTALLED:
            raise ComponentMissingDependencyError(self, _BINCOPY_TOOL)

        program_child = await resource.get_only_child_as_view(IhexProgram)
        vaddr_offset = -program_child.address_limits.start
        binfile = BinFile()
        binfile.execution_start_address = program_child.start_addr
        for seg in program_child.segments:
            seg_data = await program_child.resource.get_data(seg.translate(vaddr_offset))
            binfile.add_binary(seg_data, seg.start)

        new_data = binfile.as_ihex()
        if new_data.endswith("\n"):
            new_data = new_data[:-1]
        new_data = new_data.encode("utf-8")
        old_data_len = await resource.get_data_length()
        resource.queue_patch(Range(0, old_data_len), new_data)


class IhexIdentifier(Identifier):
    """
    Regex-test the entire resource to check if it satisfies intel-hex formatting.

    This identifier tags any Resource whose first two lines match the ihex format.
    """

    targets = (GenericText,)

    # Matches on 2 lines that match ihex format
    _INTEL_HEX_PATTERN = re.compile(rb"((\:([0-9A-F]{2}){5,})(\n|\r\n)+){2}")

    async def identify(self, resource: Resource, config=None) -> None:
        matched_ihex = await resource.search_data(self._INTEL_HEX_PATTERN, max_matches=1)
        if matched_ihex:
            offset, bytes = matched_ihex[0]
            # Only tag if pattern starts at offset 0 of resource
            if offset == 0:
                resource.add_tag(Ihex)


def _binfile_analysis(raw_ihex: bytes, component) -> Tuple[IhexProgram, Any]:
    if not BINCOPY_INSTALLED:
        raise ComponentMissingDependencyError(component, _BINCOPY_TOOL)
    binfile = BinFile()
    binfile.add_ihex(raw_ihex.decode("utf-8"))

    ihex_program = IhexProgram(
        Range(binfile.minimum_address, binfile.maximum_address),
        binfile.execution_start_address,
        [Range(segment.minimum_address, segment.maximum_address) for segment in binfile.segments],
    )
    return ihex_program, binfile



import hashlib
from dataclasses import dataclass

from ofrak.core.binary import GenericBinary
from ofrak.core.filesystem import File

from ofrak.component.analyzer import Analyzer
from ofrak.model.resource_model import ResourceAttributes
from ofrak.resource import Resource


@dataclass(**ResourceAttributes.DATACLASS_PARAMS)
class Sha256Attributes(ResourceAttributes):
    checksum: str


class Sha256Analyzer(Analyzer[None, Sha256Attributes]):
    """
    Analyze binary data and add attributes with the SHA256 checksum of the data.
    """

    targets = (File, GenericBinary)
    outputs = (Sha256Attributes,)

    async def analyze(self, resource: Resource, config=None) -> Sha256Attributes:
        data = await resource.get_data()
        sha256 = hashlib.sha256()
        sha256.update(data)
        return Sha256Attributes(sha256.hexdigest())


@dataclass(**ResourceAttributes.DATACLASS_PARAMS)
class Md5Attributes(ResourceAttributes):
    checksum: str


class Md5Analyzer(Analyzer[None, Md5Attributes]):
    """
    Analyze binary data and add attributes with the MD5 checksum of the data.
    """

    targets = (File, GenericBinary)
    outputs = (Md5Attributes,)

    async def analyze(self, resource: Resource, config=None) -> Md5Attributes:
        data = await resource.get_data()
        md5 = hashlib.md5()
        md5.update(data)
        return Md5Attributes(md5.hexdigest())



from abc import ABC, abstractmethod
from dataclasses import dataclass
from typing import Optional, Tuple, Iterable

from ofrak_type.architecture import (
    InstructionSet,
    SubInstructionSet,
    InstructionSetMode,
    ProcessorType,
)
from ofrak.service.abstract_ofrak_service import AbstractOfrakService
from ofrak_type.bit_width import BitWidth
from ofrak_type.endianness import Endianness


@dataclass
class DisassemblerServiceRequest:
    isa: InstructionSet
    sub_isa: Optional[SubInstructionSet]
    bit_width: BitWidth
    endianness: Endianness
    processor: Optional[ProcessorType]
    mode: InstructionSetMode
    data: bytes
    virtual_address: int


@dataclass
class DisassemblyResult:
    address: int
    size: int
    mnemonic: str
    operands: str


@dataclass
class RegisterUsageResult:
    regs_read: Tuple[str, ...]
    regs_written: Tuple[str, ...]


class DisassemblerServiceInterface(AbstractOfrakService, ABC):
    @abstractmethod
    async def disassemble(self, request: DisassemblerServiceRequest) -> Iterable[DisassemblyResult]:
        raise NotImplementedError

    @abstractmethod
    async def get_register_usage(self, request: DisassemblerServiceRequest) -> RegisterUsageResult:
        raise NotImplementedError


class DisassemblerArchSupportError(Exception):
    pass


class DisassemblerRegisterUsageSupportError(Exception):
    pass



from enum import Enum
from typing import Any, Type

import inspect

from ofrak.service.serialization.serializers.serializer_i import SerializerInterface


def is_enum(type_hint):
    return type_hint == Enum or (inspect.isclass(type_hint) and issubclass(type_hint, Enum))


class EnumSerializer(SerializerInterface):
    """
    Serialize and deserialize enum.Enum instances into `PJSONType`.

    Implementation: the class is serialized, along with the name of the attribute for this instance, into a string.
    """

    targets = (is_enum,)

    def obj_to_pjson(self, enum_instance: Enum, _type_hint: Any) -> str:
        return f"{self._service.to_pjson(enum_instance.__class__, Type)}.{enum_instance.name}"

    def pjson_to_obj(self, pjson_obj: str, _type_hint: Any) -> Enum:
        enum_cls_ref_pjson, enum_member_name = pjson_obj.rsplit(".", maxsplit=1)
        enum_cls = self._service.from_pjson(enum_cls_ref_pjson, Type)
        return enum_cls[enum_member_name]



from abc import ABCMeta, abstractmethod
from typing import List, Iterable, Optional, Tuple, overload, Pattern

from ofrak.model.data_model import DataModel, DataPatch, DataPatchesResult
from ofrak.service.abstract_ofrak_service import AbstractOfrakService
from ofrak_type.range import Range


class DataServiceInterface(AbstractOfrakService, metaclass=ABCMeta):
    @abstractmethod
    async def create_root(self, data_id: bytes, data: bytes) -> DataModel:
        """
        Create a root data model with its own data bytes.

        :param data_id: Unique ID for the new data model
        :param data: Binary data belonging to the new data model

        :return: The new data model object

        :raises AlreadyExistError: if `data_id` is already associated with a model
        """
        raise NotImplementedError()

    @abstractmethod
    async def create_mapped(
        self,
        data_id: bytes,
        parent_id: bytes,
        range_in_parent: Range,
    ) -> DataModel:
        """
        Create a new data model which is mapped into another data model. That is, it does not hold
        its own data, but defines its own data as a subsection of another model's data. The model
        it maps from (`parent_id`) may be a root model or another mapped model; if `parent_id` is
        another mapped node, the new mapped node created here will be mapped to the same root as
        `parent_id` at a range translated to be within `parent_id` as defined by `range_in_parent`.

        :param data_id: Unique ID for the new data model
        :param parent_id: ID of the data model to map the new model into
        :param range_in_parent: Range in `parent_id` which the new model will map

        :return: The new data model object

        :raises AlreadyExistError: if `data_id` is already associated with a model
        :raises NotFoundError: if `parent_id` is not associated with any known model
        """
        raise NotImplementedError()

    @abstractmethod
    async def get_by_id(self, data_id: bytes) -> DataModel:
        """
        Get the data model object associated with the given ID.

        :param data_id: A unique ID for a data model

        :return: The model associated with `data_id`

        :raises NotFoundError: if `data_id` is not associated with any known model
        """
        raise NotImplementedError()

    @abstractmethod
    async def get_by_ids(self, data_ids: Iterable[bytes]) -> Iterable[DataModel]:
        """
        Get the data models object associated with the given IDs.

        :param data_ids: Multiple unique IDs for data models

        :return: The models associated with each ID in `data_ids`, in the same order their IDs were
        provided

        :raises NotFoundError: if any ID in `data_ids` is not associated with any known model
        """
        raise NotImplementedError()

    @abstractmethod
    async def get_data_length(self, data_id: bytes) -> int:
        """
        Return the length of a single data model.

        :param data_id: A unique ID for a data model

        :return: The length of the data included in the model

        :raises NotFoundError: if `data_id` is not associated with any known model
        """
        raise NotImplementedError()

    @abstractmethod
    async def get_data_range_within_root(self, data_id: bytes) -> Range:
        """
        Get the range that a model maps in its root. If the model specified by `data_id` is itself
        a root, returns a range covering that whole root (i.e. Range(0, length)).

        :param data_id: A unique ID for a data model

        :return: Range that `data_id` maps in its root

        :raises NotFoundError: if `data_id` is not associated with any known model
        """
        raise NotImplementedError()

    @abstractmethod
    async def get_range_within_other(self, data_id: bytes, within_data_id: bytes) -> Range:
        """
        Get the range representing the intersection between two data models, assuming they are both
        mapped into the same root data. Either of `data_id` or `within_data_id` may be roots, but
        they cannot both be roots (unless they are the same).

        :param data_id: A unique ID for a data model
        :param within_data_id: A unique ID for a data model

        :return: The range where `data_id`'s model intersects `within_data_id`'s model

        :raises NotFoundError: if `data_id` or `within_data_id` is not associated with any known
        model
        :raises ValueError: if `data_id` is not mapped into `within_data_id` or they do not share
        the same root
        """
        raise NotImplementedError()

    @abstractmethod
    async def get_data(self, data_id: bytes, data_range: Optional[Range] = None) -> bytes:
        """
        Get the data (or section of data) of a model. The optional `data_range` parameter specifies
        which a range within `data_id`'s data to return; if this range actually falls outside the
        boundaries of `data_id`'s data, an empty bytestring is returned.

        :param data_id: A unique ID for a data model
        :param data_range: An optional range within the model's data to return

        :return: Bytes of data from the model associated with `data_id` - all bytes by default, a
        specific slice if `data_range` is provided, and empty bytes if `data_range` is provided but
        is outside the modeled data.

        :raises NotFoundError: if `data_id` is not associated with any known model
        """
        raise NotImplementedError()

    @abstractmethod
    async def apply_patches(
        self,
        patches: List[DataPatch],
    ) -> List[DataPatchesResult]:
        """
        Modify the data of a number of models, modeled as a list of `DataPatch` structures each
        specifying: a target data model (by ID), new data, and a range to overwrite with the new
        data. The listed patches are applied in order, so that subsequent patches may effectively
        'erase' an earlier patch. Patches may resize data if the new data is not the same size as
        the range it is overwriting. Such patches create additional restrictions:

        1. If `patches` contains a patch that resizes a range of data, no subsequent patch in
        `patches` is allowed to modify that resized range.
        2. Resizing patches are not allowed to overwrite ranges that contain the borders of any
        data models. For example, if model B maps Range(0, 6) of model A, a patch that resizes
        Range(4, 10) of model A is not allowed (whether it increases or decreases the size).

        :param patches: A list of patch data structures to be applied, in order

        :return: A list of data structures describing all modified ranges of each data model
        affected by `patches`

        :raises NotFoundError: if any data ID in the `patches` list is not associated with any
        known model
        :raises PatchOverlapError: if a patch targets a region of data which has already been
        modified by a patch which resized that region
        :raises PatchOverlapError: if a patch would resize a region of data which contains the
        start or end of one or more data models
        """
        raise NotImplementedError()

    @abstractmethod
    async def delete_models(self, data_ids: Iterable[bytes]) -> None:
        """
        Delete one or more data models. If a root model is deleted, all models mapped into that
        root are also deleted.

        :param data_ids: Multiple unique IDs for data models

        :raises NotFoundError: if any ID in `data_ids` is not associated with any known model
        """
        raise NotImplementedError()

    @overload
    @abstractmethod
    async def search(
        self,
        data_id: bytes,
        query: Pattern[bytes],
        start: Optional[int] = None,
        end: Optional[int] = None,
        max_matches: Optional[int] = None,
    ) -> Tuple[Tuple[int, bytes], ...]:
        ...

    @overload
    @abstractmethod
    async def search(
        self,
        data_id: bytes,
        query: bytes,
        start: Optional[int] = None,
        end: Optional[int] = None,
        max_matches: Optional[int] = None,
    ) -> Tuple[int, ...]:
        ...

    @abstractmethod
    async def search(self, data_id, query, start=None, end=None, max_matches=None):
        """
        Search for some data in one of the models. The query may be a regex pattern (a return value
        of `re.compile`). If the query is a regex pattern, returns a tuple of pairs with both the
        offset of the match and the contents of the match itself. If the query is plain bytes, a
        list of only the match offsets are returned.

        :param data_id: Data model to search
        :param query: Plain bytes to exactly match or a regex pattern to search for
        :param start: Start offset in the data model to begin searching
        :param end: End offset in the data model to stop searching
        :param max_matches: Maximum number of matches to return

        :return: A tuple of offsets matching a plain bytes query, or a list of (offset, match) pairs
        for a regex pattern query
        """
        raise NotImplementedError()



import itertools
from typing import FrozenSet, Set, Tuple, Type

from dataclasses import dataclass

from ofrak.model.resource_model import ResourceAttributes

from ofrak.model.tag_model import ResourceTag

from ofrak.component.interface import ComponentInterface

from ofrak.component.analyzer import Analyzer
from ofrak.service.component_locator_i import ComponentFilter

from functools import lru_cache


@lru_cache(maxsize=None)
def _isinstance(*args, **kwargs):
    return isinstance(*args, **kwargs)


@dataclass(frozen=True, init=False)
class ComponentWhitelistFilter(ComponentFilter):
    """
    Only allow components which belong to a specific set to be run.
    """

    whitelisted_component_ids: FrozenSet[bytes]

    def __init__(self, *whitelisted_component_ids):
        object.__setattr__(self, "whitelisted_component_ids", frozenset(whitelisted_component_ids))

    def filter(self, components: Set[ComponentInterface]) -> Set[ComponentInterface]:
        return {c for c in components if c.get_id() in self.whitelisted_component_ids}


@dataclass(frozen=True)
class ComponentTypeFilter(ComponentFilter):
    """
    Only allow components of a specific type (e.g. Analyzer) to run.
    """

    component_type: Type[ComponentInterface]

    def __repr__(self) -> str:
        return f"ComponentTypeFilter({self.component_type.__name__})"

    def filter(self, components: Set[ComponentInterface]) -> Set[ComponentInterface]:
        return {c for c in components if _isinstance(c, self.component_type)}


@dataclass(frozen=True, init=False)
class ComponentTargetFilter(ComponentFilter):
    """
    Only allow components which target at least one of the tags in a set. The tags must be
    strictly equal, that is, super/subclasses of the tags are not checked.
    """

    tags: Tuple[ResourceTag, ...]

    def __init__(self, *tags: ResourceTag):
        object.__setattr__(self, "tags", tags)

    def __repr__(self) -> str:
        return f"ComponentTargetFilter({', '.join(t.__name__ for t in self.tags)})"

    def filter(self, components: Set[ComponentInterface]) -> Set[ComponentInterface]:
        return {c for c in components if any(t in c.targets for t in self.tags)}


@dataclass(frozen=True, init=False)
class AnalyzerOutputFilter(ComponentFilter):
    """
    Only allow analyzers whose outputs have some overlap with the requested outputs.
    """

    outputs: FrozenSet[Type[ResourceAttributes]]

    def __init__(self, *outputs: Type[ResourceAttributes]):
        object.__setattr__(self, "outputs", frozenset(outputs))

    def filter(self, components: Set[ComponentInterface]) -> Set[ComponentInterface]:
        def component_is_analyzer_with_outputs(c: ComponentInterface):
            if _isinstance(c, Analyzer):
                analyzer_outputs = set(c.get_outputs_as_attribute_types())  # type: ignore
                return self.outputs.issubset(analyzer_outputs)
            return False

        return {c for c in components if component_is_analyzer_with_outputs(c)}

    def __repr__(self) -> str:
        return f"AnalyzerOutputFilter({', '.join(attr_t.__name__ for attr_t in self.outputs)})"


@dataclass(frozen=True, init=False)
class ComponentOrMetaFilter(ComponentFilter):
    """
    Only allow components which match any one of multiple filters. If there are no filters, no
    components will be filtered out.
    """

    filters: Tuple[ComponentFilter, ...]

    def __init__(self, *filters: ComponentFilter):
        object.__setattr__(self, "filters", filters)

    def __repr__(self) -> str:
        return f"({' or '.join(f.__repr__() for f in self.filters)})"

    def filter(self, components: Set[ComponentInterface]) -> Set[ComponentInterface]:
        if 0 == len(self.filters):
            return components
        else:
            return set(itertools.chain(*(f.filter(components) for f in self.filters)))


@dataclass(frozen=True, init=False)
class ComponentAndMetaFilter(ComponentFilter):
    """
    Only allow components which match all of multiple filters. If there are no filters, all
    components will be filtered out.
    """

    filters: Tuple[ComponentFilter, ...]

    def __init__(self, *filters: ComponentFilter):
        object.__setattr__(self, "filters", filters)

    def __repr__(self) -> str:
        return f"({' and '.join(f.__repr__() for f in self.filters)})"

    def filter(self, components: Set[ComponentInterface]) -> Set[ComponentInterface]:
        if 0 == len(self.filters):
            return set()
        for component_filter in self.filters:
            components = component_filter.filter(components)
            if not components:
                break

        return components


@dataclass(frozen=True, init=False)
class ComponentPrioritySelectingMetaFilter(ComponentFilter):
    """
    Selects exactly one filter to apply from a prioritized list of component filters. Only the first
    filter which allows more than zero components is applied. If no filters allow any components
    through, then this filter passes no components.

    For example, if the filters are:
    - filter 1: tag matches A, B, or C
    - filter 2: tag matches C, D, or E
    - filter 3: tag matches E, F, or G

    and the components under consideration are one targeting (E) and one targeting (F), then only
    the component targeting (E) passes this meta-filter. This is because:
    - filter 1 would filter out all components under consideration, so it's ignored;
    - filter 2 would allow one or more components (namely, the component targeting (E)), so this
    meta-filter then behaves like just filter 2 in this prioritized list.
    """

    filters: Tuple[ComponentFilter, ...]

    def __init__(self, *filters: ComponentFilter):
        object.__setattr__(self, "filters", filters)

    def __repr__(self) -> str:
        return f"({' then '.join(f.__repr__() for f in self.filters)})"

    def filter(self, components: Set[ComponentInterface]) -> Set[ComponentInterface]:
        for f in self.filters:
            components_passing_f = f.filter(components)
            if components_passing_f:
                return components_passing_f

        return set()


@dataclass(frozen=True)
class ComponentNotMetaFilter(ComponentFilter):
    """
    Invert the result of a child filter, that is, filter out components that would pass it and pass
    components which the child filter would filter out.
    """

    child_filter: ComponentFilter

    def filter(self, components: Set[ComponentInterface]) -> Set[ComponentInterface]:
        return components.difference(self.child_filter.filter(components))

    def __repr__(self) -> str:
        return f"not {self.child_filter}"



from ofrak.core.elf.model import *
from ofrak.core.elf.modifier import *
from ofrak.core.elf.unpacker import *
from ofrak.core.elf.analyzer import *
from ofrak.core.elf.arch.x86_64.model import *
from ofrak.core.elf.lief_modifier import *

from ofrak.core.pe.unpacker import *
from ofrak.core.pe.model import *

from ofrak.core.patch_maker.linkable_binary import *
from ofrak.core.patch_maker.linkable_symbol import *
from ofrak.core.patch_maker.modifiers import *

from ofrak.core.addressable import *
from ofrak.core.apk import *
from ofrak.core.architecture import *
from ofrak.core.architecture_analysis import *
from ofrak.core.basic_block import *
from ofrak.core.binary import *
from ofrak.core.binwalk import *
from ofrak.core.bzip2 import *
from ofrak.core.checksum import *
from ofrak.core.code_region import *
from ofrak.core.comments import *
from ofrak.core.complex_block import *
from ofrak.core.cpio import *
from ofrak.core.data import *
from ofrak.core.dtb import *
from ofrak.core.filesystem import *
from ofrak.core.flash import *
from ofrak.core.free_space import *
from ofrak.core.run_script_modifier import *
from ofrak.core.gzip import *
from ofrak.core.ihex import *
from ofrak.core.injector import *
from ofrak.core.instruction import *
from ofrak.core.iso9660 import *
from ofrak.core.label import *
from ofrak.core.lzma import *
from ofrak.core.lzo import *
from ofrak.core.magic import *
from ofrak.core.memory_region import *
from ofrak.core.openwrt import *
from ofrak.core.seven_zip import *
from ofrak.core.program import *
from ofrak.core.program_section import *
from ofrak.core.rar import *
from ofrak.core.squashfs import *
from ofrak.core.strings import *
from ofrak.core.strings_analyzer import *
from ofrak.core.tar import *
from ofrak.core.ubi import *
from ofrak.core.ubifs import *
from ofrak.core.uf2 import *
from ofrak.core.uimage import *
from ofrak.core.zip import *
from ofrak.core.zlib import *
from ofrak.core.zstd import *



from abc import ABC, abstractmethod
from typing import Optional, Tuple, Iterable

from ofrak.resource import Resource

from ofrak.component.analyzer import Analyzer
from ofrak.core.code_region import CodeRegion
from ofrak.core.complex_block import ComplexBlock
from ofrak.core.data import ReferencedStringsAttributes, ReferencedDataAttributes
from ofrak.core.memory_region import MemoryRegion
from ofrak.service.resource_service_i import (
    ResourceFilter,
    ResourceAttributeValueFilter,
    ResourceSort,
    ResourceSortDirection,
)
from ofrak.core.patch_maker.linkable_binary import LinkableBinary
from ofrak_type.error import NotFoundError
from ofrak_type.range import Range


class Program(LinkableBinary):
    """
    Generic representation for a binary program with functions, code and
    [memory regions][ofrak.core.memory_region.MemoryRegion], and
    [ProgramAttributes][ofrak.core.architecture.ProgramAttributes].
    """

    async def get_function_complex_block(self, func_name: str) -> ComplexBlock:
        return await self.resource.get_only_descendant_as_view(
            ComplexBlock,
            r_filter=ResourceFilter(
                tags=(ComplexBlock,),
                attribute_filters=(ResourceAttributeValueFilter(ComplexBlock.Symbol, func_name),),
            ),
        )

    async def get_code_regions(self) -> Iterable[CodeRegion]:
        """
        Return code regions associated with the resource

        :return: Iterable of CodeRegions
        """
        return await self.resource.get_children_as_view(
            CodeRegion, r_filter=ResourceFilter.with_tags(CodeRegion)
        )

    async def get_code_region_for_vaddr(self, vaddr: int) -> Optional[CodeRegion]:
        """
        Return the code region in this program containing vaddr.

        :param vaddr: Virtual address

        :raises NotFoundError: If vaddr is not in any code region

        :return: Code region containing the input vaddr
        """

        code_regions = await self.get_code_regions()

        for cr_view in code_regions:
            code_region_vaddr_range = Range(
                cr_view.virtual_address,
                cr_view.virtual_address + cr_view.size,
            )
            if vaddr in code_region_vaddr_range:
                return cr_view

        raise NotFoundError

    async def get_memory_region_for_vaddr(self, vaddr: int) -> Optional[MemoryRegion]:
        """
        Return the largest [memory region][ofrak.core.memory_region.MemoryRegion] containing vaddr.

        :param vaddr: Virtual address

        :raises NotFoundError: If vaddr is not in any memory region

        :return: The most general (largest) region of memory containing the input vaddr,
        if such a memory region exists
        """
        # we're looking for the largest (most general) memory region containing this vaddr
        mem_regions = await self.resource.get_descendants_as_view(
            MemoryRegion,
            r_filter=ResourceFilter.with_tags(MemoryRegion),
            r_sort=ResourceSort(
                attribute=MemoryRegion.Size,
                direction=ResourceSortDirection.DESCENDANT,
            ),
        )

        return MemoryRegion.get_mem_region_with_vaddr_from_sorted(vaddr, mem_regions)


class ReferencedStringsAnalyzer(Analyzer[None, Tuple[ReferencedStringsAttributes]], ABC):
    targets = (Program,)
    outputs = (ReferencedStringsAttributes,)

    @abstractmethod
    async def analyze(self, resource: Resource, config=None) -> Tuple[ReferencedStringsAttributes]:
        raise NotImplementedError()


class ReferencedDataAnalyzer(Analyzer[None, Tuple[ReferencedDataAttributes]], ABC):
    targets = (Program,)
    outputs = (ReferencedDataAttributes,)

    @abstractmethod
    async def analyze(self, resource: Resource, config=None) -> Tuple[ReferencedDataAttributes]:
        raise NotImplementedError()



from typing import Union, Any

from ofrak.service.serialization.serializers.serializer_i import SerializerInterface


class BasicTypeSerializer(SerializerInterface):
    """
    Serialize and deserialize basic types (which don't need any change) into `PJSONType`.
    """

    targets = (int, float, bool, str, type(None))

    BasicType = Union[int, float, bool, str, None]

    def obj_to_pjson(self, obj: BasicType, _type_hint: Any) -> BasicType:
        return obj

    def pjson_to_obj(self, pjson_obj: BasicType, _type_hint: Any) -> BasicType:
        return pjson_obj



import tempfile
from dataclasses import dataclass
from typing import List, Optional

import lief

from ofrak.component.modifier import Modifier
from ofrak.model.component_model import ComponentConfig
from ofrak.resource import Resource
from ofrak.core.elf.model import Elf
from ofrak_type.range import Range


@dataclass
class LiefAddSegmentConfig(ComponentConfig):
    """
    Config for the [LiefAddSegmentModifier][ofrak.core.elf.lief_modifier.LiefAddSegmentModifier].

    :ivar virtual_address: virtual address of the new segment
    :ivar alignment: alignment of the new segment
    :ivar content: list of integers representing the raw bytes of the new segment
    :ivar rwx_flags: string representation of the new segment's R/W/X permissions
    :ivar replace_note: replace the unused NOTE segment with the new segment, rather than adding a
    new segment. defaults to `True`, as adding a new segment may corrupt the ELF due to a LIEF bug.
    :ivar physical_address: overwrite the default physical address (defaults to the virtual address)
    """

    virtual_address: int
    alignment: int
    content: List[int]
    rwx_flags: str
    replace_note: bool = True
    physical_address: Optional[int] = None


class LiefAddSegmentModifier(Modifier[LiefAddSegmentConfig]):
    id = b"LiefAddSegmentModifier"
    targets = (Elf,)

    async def modify(self, resource: Resource, config: LiefAddSegmentConfig) -> None:
        binary: lief.ELF.Binary = lief.parse(await resource.get_data())

        segment = lief.ELF.Segment()
        segment.type = lief.ELF.SEGMENT_TYPES.LOAD
        segment.content = config.content
        segment.alignment = config.alignment
        segment.virtual_address = config.virtual_address
        if config.physical_address is not None:
            segment.physical_address = config.physical_address
        if "r" in config.rwx_flags:
            segment.add(lief.ELF.SEGMENT_FLAGS.R)
        if "w" in config.rwx_flags:
            segment.add(lief.ELF.SEGMENT_FLAGS.W)
        if "x" in config.rwx_flags:
            segment.add(lief.ELF.SEGMENT_FLAGS.X)

        if config.replace_note:
            # instead of adding a segment to the binary, replace a useless NOTE segment
            #   see https://github.com/lief-project/LIEF/issues/98
            #   and https://github.com/lief-project/LIEF/issues/143
            if not binary.has(lief.ELF.SEGMENT_TYPES.NOTE):
                raise ValueError("Binary must have a NOTE section to add a new section")
            segment = binary.replace(segment, binary[lief.ELF.SEGMENT_TYPES.NOTE])
            if config.physical_address is not None:
                segment.physical_address = config.physical_address
        else:
            _ = binary.add(segment)

        with tempfile.NamedTemporaryFile() as temp_file:
            binary.write(temp_file.name)
            temp_file.flush()
            with open(temp_file.name, "rb") as f_handle:
                new_data = f_handle.read()
        # replace all old content (old range) with new content from Lief
        resource.queue_patch(Range(0, await resource.get_data_length()), new_data)



import dataclasses
from _warnings import warn
from collections import defaultdict
from dataclasses import dataclass
from typing import Tuple, Type, Dict, Any, List, Set, TypeVar, Iterable, MutableMapping, Generic

import ofrak.model._auto_attributes
from ofrak.model.resource_model import (
    ResourceAttributes,
    ResourceIndexedAttribute,
    ResourceModel,
    ResourceContext,
)
from ofrak.model.tag_model import ResourceTag

_VIEW_ATTRIBUTES_TYPE = "__view_attributes_type__"
_COMPOSED_ATTRIBUTES_TYPE = "__composed_attributes_types__"


RA = TypeVar("RA", bound=ResourceAttributes)
RVI = TypeVar("RVI", bound="ResourceViewInterface")


class AttributesType(ResourceAttributes, Generic[RVI]):
    """
    A Generic type for `ViewableResourceTag` to get the auto-generated `ResourceAttributes`
    class associated with that view type. The returned class is a `dataclass` which encapsulates
    the fields defined in one specific `ViewableResourceTag`.

    For example if `B` inherits from `A` and `A` defines one or more new fields,
    `AttributesType[B]` has only fields defined in `B`, and none of the fields defined in `A`.
    """

    def __init__(self, *args: Any, **kwargs: Any):
        raise NotImplementedError("AttributesType cannot be instantiated")

    def __class_getitem__(cls, item: Type[RVI]) -> Type[ResourceAttributes]:
        return getattr(item, _VIEW_ATTRIBUTES_TYPE)

    def __getattr__(self, item):  # pragma: no cover
        """
        Type stub that 'tricks' MyPy into not trying to typecheck attribute accesses of an
        `AttributesType` instance. This stub solves the following case:

        ```
        x: AttributesType[X] = func()
        x.any_field  <--- Always is a type error because MyPy thinks AttributesType has no fields!
        ```

        Without this stub, each instance of x.any_field would need to be marked with # type: ignore
        """
        ...


@dataclass
class _NoResourceAttributesType(ResourceAttributes):
    """
    Dummy class indicating a
    [ViewableResourceTag][ofrak.model.viewable_tag_model.ViewableResourceTag] has no
    `attributes_type` attribute.
    """


_NO_RESOURCE_ATTRIBUTES_TYPE = _NoResourceAttributesType


class ViewableResourceTag(ResourceTag):
    def __new__(mcs, name: str, bases: Tuple[Type, ...], namespace: Dict[str, Any]):
        """
        Create a new attributes type for the `ViewableResourceTag` which is about to be created.
        This type will inherit from `ResourceAttributes` and have fields corresponding to all the
        fields of unique to the new `ViewableResourceTag`. This type, as well as all attributes
        types inherited from the bases, are added to the namespace of the new class under the
        dunders `_VIEW_ATTRIBUTES_TYPE` and `_COMPOSED_ATTRIBUTES_TYPE`.

        This method also checks for possible attempted polymorphism, raising warnings if any
        method overrides are detected.

        The `__new__` method can inspect and edit the names, bases, and namespace (attributes and
        methods) of the class. It is called after the class definition has been parsed, and before
        the new class object has been created.

        :param name: Name of the new class
        :param bases: Base classes of the new class
        :param namespace: Namespace of attributes for the new class, in the form of a dictionary
        mapping attribute names to objects
        :return:
        """
        _check_for_polymorphism(name, bases, namespace)
        attributes_type = mcs._create_attributes_type(name, bases, namespace)
        composed_attributes_types = [attributes_type]
        for base_cls in bases:
            composed_attributes_types.extend(_get_attributes_types_recursively(base_cls))
        composed_attributes_types = _filter_attributes_types(composed_attributes_types)

        namespace[_VIEW_ATTRIBUTES_TYPE] = attributes_type
        namespace[_COMPOSED_ATTRIBUTES_TYPE] = composed_attributes_types
        return super().__new__(mcs, name, bases, namespace)

    def __init__(cls, name: str, bases: Tuple[Type, ...], namespace: Dict[str, Any]):
        """
        Fix up any `@index` defined in this `ViewableResourceTag`. Each index descriptor needs an
        owner which is a `ResourceAttributes` subclass, and an automatically populated owner will be
        set to the newly created `cls`, which is an instance of `ViewableResourceTag` and not a
        `ResourceAttributes` subclass. This is a little hacky, but it works just fine.

        Can inspect and edit the names, bases, and namespace (attributes and methods) of the class.

        Called after the new class object has been created.

        :param name: Name of the new class
        :param bases: Base classes of the new class
        :param namespace: Namespace of attributes for the new class, in the form of a dictionary
        mapping attribute names to objects

        :raises TypeError: If the viewable tag has no attributes but does have an @index

        :return:
        """
        # Change owner of the indexes to be the attributes type
        for name, index_descriptor in _get_indexes(namespace).items():
            if getattr(cls, _VIEW_ATTRIBUTES_TYPE) is None:
                raise TypeError(
                    f"Cannot have an index in a ResourceView which has no attributes "
                    f"- an index should only access one set of attributes, "
                    f"so this index is likely illegal anyway."
                )
            index_descriptor.__set_name__(getattr(cls, _VIEW_ATTRIBUTES_TYPE), name)

        super().__init__(cls, name, bases)  # type: ignore

    @property  # pragma: no cover
    def attributes_type(cls) -> Type[ResourceAttributes]:
        """
        Get the auto-generated `ResourceAttributes` subclass for this `ViewableResourceTag`. The
        returned class is a `dataclass` which encapsulates the fields defined in one specific
        `ViewableResourceTag`. For example if `B` inherits from `A` and `A` defines several fields,
        `B.attributes_type` has only fields defined in `B`, and none of the fields defined in `A`.

        :return: The auto-generated `ResourceAttributes` subclass for this `ViewableResourceTag`
        class, in no particular order.
        """
        warn(
            "T.attributes_type is deprecated! Use AttributesType[T] instead.",
            category=DeprecationWarning,
        )
        return getattr(cls, _VIEW_ATTRIBUTES_TYPE)

    @property
    def composed_attributes_types(cls) -> Iterable[Type[ResourceAttributes]]:
        """
        Get all of the `ResourceAttributes` subclasses which this class is composed of. This means
        walking back through the class hierarchy and getting the `AttributesType[base]` for every
        base class of this class.

        :return: The `attributes_type` of every `ViewableResourceTag` this class inherits from,
        in no particular order.
        """
        return getattr(cls, _COMPOSED_ATTRIBUTES_TYPE)

    @classmethod
    def _create_attributes_type(
        mcs, name: str, bases: Tuple[Type, ...], namespace: Dict[str, Any]
    ) -> Type[ResourceAttributes]:
        """
        Get a type inheriting from `ResourceAttributes` which is unique to this tag (`cls`).
        :return:
        """

        # First make this a dataclass to easily get its fields
        # We can't depend on the class already being a dataclass because class decorators run
        # after metaclass __new__ and __init__ methods, so it is not yet a dataclass
        tmp_cls = super().__new__(mcs, name, bases, namespace)
        tmp_dataclass: object = dataclass(tmp_cls)  # type: ignore

        base_fields: Set[dataclasses.Field] = set()
        for base in bases:
            if type(base) is ViewableResourceTag:
                base_fields.update(dataclasses.fields(base))

        fields = [
            (_field.name, _field.type, _field)
            for _field in dataclasses.fields(tmp_dataclass)
            if _field not in base_fields and not _field.name.startswith("_")
        ]
        if len(fields) == 0:
            return _NO_RESOURCE_ATTRIBUTES_TYPE
        indexed_attributes_namespace = _get_indexes(namespace)
        # Creates a new class inheriting from ResourceAttributes, with the same fields as this
        # ViewableResourceTag, as well as the same indexed attribute descriptors
        attributes_type = dataclasses.make_dataclass(
            f"{AttributesType.__name__}[{name}]",
            fields,
            bases=(ResourceAttributes,),
            namespace=indexed_attributes_namespace,
            **ResourceAttributes.DATACLASS_PARAMS,
        )

        # By default, this new attributes_type is part of the "types" module.
        # Instead, we make it part of the module ofrak.model._auto_attributes.
        attributes_type.__module__ = "ofrak.model._auto_attributes"
        setattr(
            ofrak.model._auto_attributes,
            attributes_type.__name__,
            attributes_type,
        )

        return attributes_type


def _get_attributes_types_recursively(cls: type) -> List[Type[ResourceAttributes]]:
    if isinstance(cls, ViewableResourceTag):
        attrs_types = [getattr(cls, _VIEW_ATTRIBUTES_TYPE)]
        for base_cls in cls.__bases__:
            attrs_types.extend(_get_attributes_types_recursively(base_cls))
        return attrs_types
    else:
        return []


def _filter_attributes_types(
    attr_types: List[Type[ResourceAttributes]],
) -> List[Type[ResourceAttributes]]:
    filtered_types = set()
    for attr_type in attr_types:
        if attr_type is not _NO_RESOURCE_ATTRIBUTES_TYPE:
            filtered_types.add(attr_type)
    return list(filtered_types)


def _check_for_polymorphism(name: str, bases: Iterable[Type], namespace: Dict[str, Any]):
    """
    Check for any methods in a new class which override a parent's method. The behavior of
    `view_as` means that overriding methods might not work as users think it does. Calling
    something like `a = resource.view_as(A)` will always and only return instances of `A`. These
    resources may have tags `B` and/or `C` which inherit from `A` but the returned view is not an
    instance of `B` or `C`. Calling a method `a.foo()` on that resource will therefore **ALWAYS**
    dispatch to `A.foo`, never to `B.foo` or `C.foo`.

    This would break a design in which a user expects to only program against a common viewable
    tag's interface and let individual instances dispatch to some unique behavior. For example, if
    one gets all descendants with a common tag as a view of that tag, then calls a virtual method
    of those views, expecting each view to possibly do some unique behavior. Such a class hierarchy
    could be made to work, but we raise a warning to ensure that developers are conscious of this
    restriction and program accordingly.

    :param name: Name of the new class
    :param bases: Base classes of the new class
    :param namespace: Namespace of attributes for the new class, in the form of a dictionary
    mapping attribute names to objects
    """
    for base_cls in bases:
        # TODO: Figure out cleaner way to make an exception for ResourceView
        if name == "ResourceView":
            continue
        if isinstance(base_cls, ViewableResourceTag):
            parent_cls = base_cls

            parent_class_namespace = dir(parent_cls)

            common_attrs = set(namespace.keys()).intersection(set(parent_class_namespace))
            common_attrs = common_attrs.difference(dir(ViewableResourceTag))

            common_methods = {
                attr_name: namespace[attr_name]
                for attr_name in common_attrs
                if callable(namespace[attr_name])
            }

            overwritten_methods = {
                attr_name: attr
                for attr_name, attr in common_methods.items()
                if namespace[attr_name] != getattr(parent_cls, attr_name)
            }

            for method_name, method in overwritten_methods.items():
                warn(
                    f"{name}.{method_name} overrides the parent's method "
                    f"{method_name}; OFRAK Resources do not support runtime polymorphism, "
                    f"and this function may depend on runtime polymorphism."
                )


def _get_indexes(namespace: Dict[str, Any]) -> Dict[str, ResourceIndexedAttribute]:
    """
    Extract the index descriptors from a namespaces.
    """
    return {
        name: item for name, item in namespace.items() if isinstance(item, ResourceIndexedAttribute)
    }


@dataclass
class ResourceViewInterface(metaclass=ViewableResourceTag):
    def get_attributes_instances(self) -> Dict[Type[RA], RA]:  # pragma: no cover
        raise NotImplementedError()

    def set_deleted(self):  # pragma: no cover
        """
        Mark that the underlying resource has been deleted.
        :return:
        """
        raise NotImplementedError()

    @classmethod  # pragma: no cover
    def create(cls: Type[RVI], resource_model: ResourceModel) -> RVI:  # pragma: no cover
        raise NotImplementedError()


class ResourceViewContext:
    ViewByTag = MutableMapping[ViewableResourceTag, ResourceViewInterface]

    def __init__(self):
        self.views_by_resource: MutableMapping[bytes, ResourceViewContext.ViewByTag] = defaultdict(
            dict
        )

    def has_view(self, resource_id: bytes, view_type: ViewableResourceTag) -> bool:
        return view_type in self.views_by_resource[resource_id]

    def get_view(self, resource_id: bytes, view_type: Type[RVI]) -> RVI:
        return self.views_by_resource[resource_id][view_type]  # type: ignore

    def add_view(self, resource_id: bytes, view: ResourceViewInterface):
        self.views_by_resource[resource_id][type(view)] = view

    def update_views(
        self, modified: Iterable[bytes], deleted: Iterable[bytes], resource_context: ResourceContext
    ):
        for resource_id in modified:
            views_in_context = self.views_by_resource[resource_id]
            for view in views_in_context.values():
                updated_model = resource_context.resource_models[resource_id]
                fresh_view = view.create(updated_model)
                for field in dataclasses.fields(fresh_view):
                    if field.name == "_resource":
                        continue
                    setattr(view, field.name, getattr(fresh_view, field.name))

        for resource_id in deleted:
            views_in_context = self.views_by_resource[resource_id]
            for view in views_in_context.values():
                view.set_deleted()



from abc import ABC
from typing import Optional

import pytest

from ofrak import OFRAKContext
from ofrak.model.tag_model import ResourceTag
from ofrak.resource import Resource
from ofrak.core.strings import StringPatchingConfig, StringPatchingModifier
from pytest_ofrak.patterns.unpack_modify_pack import UnpackModifyPackPattern


class CompressedFileUnpackModifyPackPattern(UnpackModifyPackPattern, ABC):
    """
    Generic test pattern for unpackers/packers targeting compressed file types. The test expects
    to create an initial compressed file with some known data, unpack it, modify the unpacked
    child by changing the string data, repack it, then verify the repacked file by independently
    decompressing it and asserting the data is equal to the expected repacked string.

    Subclasses should implement:

    :async def verify(self, repacked_root_resource: Resource):

    :def create_test_file(self):
    """

    INITIAL_DATA: bytes = b"hello world\n"
    EXPECTED_REPACKED_DATA: bytes = b"hello ofrak\n"

    @property
    def expected_tag(self) -> Optional[ResourceTag]:
        """
        Optionally check that the root_resource has a specific flag after the `unpack` method is
        run. In other words, if ``expected_tag`` is set then this will also test that the root
        resource is automatically identified.
        Set ``expected_tag`` like this:

        class TestExampleFilesystem(CompressedFileUnpackModifyPackPattern):
            expected_tag = ExampleTag

            ...

        Or omit it to skip this check.
        """
        return None

    @pytest.fixture(autouse=True)
    def create_test_file(self):
        """
        Override this method with a fixture which creates the initial compressed file to unpack.
        The file should be compressed with the relevant compression scheme and its uncompressed
        data should be CompressedFileUnpackModifyPackPattern.INITIAL_DATA
        :return:
        """
        self._test_file = None

    async def create_root_resource(self, ofrak_context: OFRAKContext) -> Resource:
        assert self._test_file
        return await ofrak_context.create_root_resource_from_file(self._test_file)

    async def unpack(self, root_resource: Resource):
        await root_resource.unpack()
        if self.expected_tag:
            assert root_resource.has_tag(self.expected_tag)

    async def repack(self, modified_root_resource: Resource):
        await modified_root_resource.pack()

        children = await modified_root_resource.get_children()
        assert 0 == len(list(children))

    async def modify(self, unpacked_root_resource: Resource):
        resource_to_modify = await unpacked_root_resource.get_only_child()
        new_string_config = StringPatchingConfig(6, "ofrak")
        await resource_to_modify.run(StringPatchingModifier, new_string_config)



from dataclasses import dataclass, field
from enum import IntEnum
import logging
import re
import os
from typing import Any, Dict, List, Optional, Tuple
from ofrak.model.resource_model import ResourceIndexedAttribute
from ofrak.resource import Resource
from ofrak.core.dtb import DtbNode, DtbProperty
from ofrak.core import Addressable
from ofrak.core.filesystem import FilesystemEntry
from ofrak.model.resource_model import Data
from ofrak.service.resource_service_i import ResourceAttributeValueFilter, ResourceFilter
from black import format_str, FileMode


LOGGER = logging.getLogger(__name__)


class SelectableAttributesError(Exception):
    """
    Prompt the user for an attribute to select with
    """


class ActionType(IntEnum):
    UNPACK = 0
    MOD = 1
    PACK = 2
    UNDEF = 3


@dataclass(frozen=True)
class ScriptAction:
    """
    Encapsulates the structure of a single action within the script, which consists of the string
    representation of the code for that action and the action's type.
    """

    action_type: ActionType
    action: str


@dataclass
class ScriptSession:
    """
    A script, consisting of an ordered sequence of script actions and a mapping between resources
    and their autogenerated variable names.
    """

    actions_queue: List[ScriptAction] = field(default_factory=list)
    actions: List[ScriptAction] = field(default_factory=list)
    resource_variable_names: Dict[bytes, str] = field(default_factory=dict)
    resource_variable_names_queue: Dict[bytes, str] = field(default_factory=dict)

    boilerplate_header: str = r"""
    from ofrak import *
    from ofrak.core import *

    async def main(ofrak_context: OFRAKContext, root_resource: Optional[Resource] = None):"""
    # TODO: Replace with backend in use by OFRAK instance used to create the script.
    boilerplate_footer: str = r"""
    if __name__ == "__main__":
        ofrak = OFRAK()
        if False:
            import ofrak_angr
            import ofrak_capstone

            ofrak.discover(ofrak_capstone)
            ofrak.discover(ofrak_angr)

        if False:
            import ofrak_binary_ninja
            import ofrak_capstone

            ofrak.discover(ofrak_capstone)
            ofrak.discover(ofrak_binary_ninja)

        if False:
            import ofrak_ghidra

            ofrak.discover(ofrak_ghidra)

        ofrak.run(main)
    """

    def get_var_name(self, id: bytes) -> str:
        if id in self.resource_variable_names:
            return self.resource_variable_names[id]
        elif id in self.resource_variable_names_queue:
            return self.resource_variable_names_queue[id]
        else:
            raise ValueError(f"No variable name for resource ID 0x{id.hex()}")


class ScriptBuilder:
    """
    Builds and maintains runnable OFRAK scripts as sequences of actions, with each script tied to
    a session.
    """

    def __init__(self):
        self.root_cache: Dict[bytes, Resource] = {}
        self.script_sessions: Dict[bytes, ScriptSession] = {}
        self.selectable_indexes: List[ResourceIndexedAttribute] = [
            FilesystemEntry.Name,
            Addressable.VirtualAddress,
            Data.Offset,
            DtbNode.DtbNodeName,
            DtbProperty.DtbPropertyName,
        ]

    async def add_action(
        self,
        resource: Resource,
        action: str,
        action_type: ActionType,
    ) -> None:
        """
        Adds an action to the script session queue to which the selected resource belongs. An action
        is a string representing the code that is being run on the resource based on an action that
        has occurred in the GUI.

        Actions are queued so that invalid actions which result in runtime exceptions within OFRAK
        do not make it into the final script. Once an action is queued and the corresponding OFRAK
        calls have run, the caller must explicitly call `clear_script_queue` or `commit_to_script`
        depending on whether an exception was raised or not, respectively.

        :param resource: Resource upon which the action is being taken
        :param action: A string describing the code being run based on a GUI action
        :param action_type: An instance of `ActionType` categorizing the action
        """
        var_name = await self._add_variable(resource)
        qualified_action = action.format(resource=var_name)
        await self._add_action_to_session_queue(resource, qualified_action, action_type)

    async def get_script(self, resource: Resource) -> List[str]:
        """
        Returns the most up-to-date version of the script for the session to which the resource
        belongs.

        :param resource: Resource belonging to the session for which the script is to be returned

        :return: List of strings where each entry is a line in the script
        """
        root_resource = await self._get_root_resource(resource)
        return self._get_script(root_resource.get_id())

    async def commit_to_script(self, resource: Resource) -> None:
        """
        Commits the staged actions and variable names in the queue to the script session following
        a one or more valid actions being run.

        :param resource: Resource belonging to the session whose queue will be committed
        """
        root_resource = await self._get_root_resource(resource)
        session = self._get_session(root_resource.get_id())
        for id, name in session.resource_variable_names_queue.items():
            session.resource_variable_names[id] = name
        session.actions += session.actions_queue
        session.actions_queue = []
        session.resource_variable_names_queue = {}

    async def clear_script_queue(self, resource: Resource) -> None:
        """
        Clears the script session queue of all staged actions and variable names following an
        invalid action being run.

        :param resource: Resource belonging to the session whose queue will be cleared
        """
        root_resource = await self._get_root_resource(resource)
        session = self._get_session(root_resource.get_id())
        session.actions_queue = []
        session.resource_variable_names_queue = {}

    async def _add_variable(self, resource: Resource) -> str:
        """
        Replaces references to a particular resource selected in the GUI with a generated variable
        name based on uniquely identifying characteristics of the resource. This overcomes the issue
        of referencing the same resource across OFRAK contexts due to the randomly generated
        resource IDs changing.

        :param resource: Resource that needs to be uniquely identified in the script

        :return: a unique variable name
        """
        if await self._var_exists(resource):
            return await self._get_variable_from_session(resource)

        root_resource = await self._get_root_resource(resource)
        if resource.get_id() == root_resource.get_id():
            await self._add_variable_to_session_queue(resource, "root_resource")
            return "root_resource"

        parent = await resource.get_parent()
        if not await self._var_exists(parent):
            await self._add_variable(parent)

        name = ""
        # Cannot propagate exceptions to the server as this would interfere with user actions
        # regardless of whether they're interested in the script. Currently only _get_selector()
        # and _generate_name() can lead to exceptions raised within ScriptBuilder.
        try:
            selector = await self._get_selector(resource)
            name = await self._generate_name(resource)
            await self._add_action_to_session_queue(
                resource,
                rf"""
        {name} = {selector}""",
                ActionType.UNDEF,
            )
            await self._add_variable_to_session_queue(resource, name)
        except SelectableAttributesError as e:
            name = await self._generate_missing_name(resource, e)
            LOGGER.exception("Could not find selectable attributes for resource")
            return name
        except:
            LOGGER.exception("Exception raised in add_variable")
        return name

    async def _generate_missing_name(self, resource: Resource, e: Exception) -> str:
        root_resource = await self._get_root_resource(resource)
        session = self._get_session(root_resource.get_id())
        parent = await resource.get_parent()
        parent_name = await self._get_variable_from_session(parent)
        name = f"{parent_name}_MISSING_RESOURCE_0"
        index = 1
        var_names = list(session.resource_variable_names.values()) + list(
            session.resource_variable_names_queue.values()
        )
        name = self._increment_missing_name_index(name, var_names, index)
        await self._add_action_to_session_queue(
            resource,
            f"""
        # Resource with parent {parent_name} is missing, could not find selectable attributes.
        raise RuntimeError(\"{str(e)}\")
        {name} = None""",
            ActionType.UNDEF,
        )
        await self._add_variable_to_session_queue(resource, name)
        return name

    def _increment_missing_name_index(self, name: str, var_names: List[str], index: int) -> str:
        name_template = "_".join(name.split("_")[:-1])
        if name in var_names:
            name = name_template + f"_{index}"
            index += 1
            return self._increment_missing_name_index(name, var_names, index)
        else:
            return name

    async def _get_root_resource(self, resource: Resource) -> Resource:
        """
        Maps a given resource to its root for efficient retrieval of the root resource because
        getting the root resource is likely the most performed operation in `ScriptBuilder`.
        """
        resource_id = resource.get_id()
        if resource_id in self.root_cache:
            return self.root_cache[resource_id]
        ancestors = list(await resource.get_ancestors())
        root = ancestors[-1] if ancestors else resource
        self.root_cache[resource_id] = root
        return root

    async def _get_variable_from_session(self, resource: Resource) -> str:
        root_resource = await self._get_root_resource(resource)
        session = self._get_session(root_resource.get_id())
        return session.get_var_name(resource.get_id())

    async def _var_exists(self, resource: Resource) -> bool:
        root_resource = await self._get_root_resource(resource)
        session = self._get_session(root_resource.get_id())
        return (
            resource.get_id() in session.resource_variable_names
            or resource.get_id() in session.resource_variable_names_queue
        )

    async def _add_action_to_session_queue(
        self, resource: Resource, action: str, action_type: ActionType
    ) -> None:
        root_resource = await self._get_root_resource(resource)
        session = self._get_session(root_resource.get_id())
        session.actions_queue.append(ScriptAction(action_type, action))

    async def _add_variable_to_session_queue(self, resource: Resource, var_name: str) -> None:
        root_resource = await self._get_root_resource(resource)
        session = self._get_session(root_resource.get_id())
        session.resource_variable_names_queue[resource.get_id()] = var_name

    async def _test_selectable_attributes(
        self,
        ancestor: Resource,
        resource: Resource,
        attribute: ResourceIndexedAttribute,
        attribute_value: Any,
    ) -> None:
        try:
            result = await ancestor.get_only_child(
                r_filter=ResourceFilter(
                    attribute_filters=[
                        ResourceAttributeValueFilter(attribute=attribute, value=attribute_value)
                    ],
                )
            )
        except Exception as e:
            raise SelectableAttributesError(
                f"Resource with ID 0x{resource.get_id().hex()} cannot be uniquely identified by attribute "
                f"{attribute.__name__} (resource has value {attribute_value})."
            )

    async def _get_selector(self, resource: Resource) -> str:
        root_resource = await self._get_root_resource(resource)
        session = self._get_session(root_resource.get_id())
        parent = await resource.get_parent()
        attribute, attribute_value = await self._get_selectable_attribute(resource)
        await self._test_selectable_attributes(parent, resource, attribute, attribute_value)

        attribute_value = f"{attribute_value!r}"
        return f"""await {session.get_var_name(parent.get_id())}.get_only_child(
                    r_filter=ResourceFilter(
                        tags={resource.get_most_specific_tags()},
                        attribute_filters=[
                            ResourceAttributeValueFilter(
                                attribute={attribute.__name__}, 
                                value={attribute_value}
                            )
                        ]   
                    )
                )"""

    async def _get_self_and_siblings_matching_attribute(
        self, resource: Resource, attribute: ResourceIndexedAttribute, attribute_value: Any
    ) -> List[Resource]:
        parent = await resource.get_parent()
        children = list(
            await parent.get_children(
                r_filter=ResourceFilter(
                    attribute_filters=[
                        ResourceAttributeValueFilter(attribute=attribute, value=attribute_value)
                    ],
                )
            )
        )
        return children

    async def _get_selectable_attribute(
        self, resource: Resource
    ) -> Tuple[ResourceIndexedAttribute, Any]:
        attribute_collisions = {}
        for attribute in self.selectable_indexes:
            if attribute.attributes_owner and resource.has_attributes(attribute.attributes_owner):
                attribute_value = attribute.get_value(resource.get_model())
                children = await self._get_self_and_siblings_matching_attribute(
                    resource, attribute, attribute_value
                )
                if len(children) == 1:
                    return attribute, attribute_value
                elif len(children) > 1:
                    attribute_collisions[attribute.__name__] = attribute_value
                    continue
        if len(attribute_collisions) == 0:
            raise SelectableAttributesError(
                f"Resource with ID 0x{resource.get_id().hex()} does not have a selectable attribute."
            )
        else:
            msg = []
            for collision, value in attribute_collisions.items():
                msg.append(
                    f"Resource with ID 0x{resource.get_id().hex()} cannot be uniquely identified by attribute {collision} (resource has value {value})."
                )
            raise SelectableAttributesError("\n".join(msg))

    async def _generate_name(self, resource: Resource) -> str:
        root_resource = await self._get_root_resource(resource)
        session = self._get_session(root_resource.get_id())
        most_specific_tag = list(resource.get_most_specific_tags())[0].__name__.lower()
        _, selectable_attribute_value = await self._get_selectable_attribute(resource)
        if isinstance(selectable_attribute_value, int):
            selectable_attribute_value = hex(selectable_attribute_value)
        name = f"{most_specific_tag}_{selectable_attribute_value}"
        name = re.sub(r"[^a-zA-Z0-9]", "_", name)
        if name in session.resource_variable_names.values():
            parent = await resource.get_parent()
            return f"{session.get_var_name(parent.get_id())}_{name}"
        return name

    def _get_session(self, resource_id: bytes) -> ScriptSession:
        session = self.script_sessions.get(resource_id, None)
        if session is None:
            session = ScriptSession()
            self.script_sessions[resource_id] = session

        return session

    def _get_script(
        self, resource_id: bytes, target_type: Optional[ActionType] = None
    ) -> List[str]:
        script_list: List[str] = []
        script_list.append(self.script_sessions[resource_id].boilerplate_header)
        for script_action in self.script_sessions[resource_id].actions:
            # Always include UNDEF actions like variable assignments
            if (
                target_type is None
                or script_action.action_type == target_type
                or script_action.action_type == ActionType.UNDEF
            ):
                script_list.append(script_action.action)
        script_list.append(self.script_sessions[resource_id].boilerplate_footer)
        script_str = "\n".join(script_list)
        script_str = self._dedent(script_str)
        try:
            res = format_str(script_str, mode=FileMode())
            script_list = res.split("\n")
        except Exception as e:
            logging.exception("Black Formatting Error:")
            logging.exception(e)
            script_list = script_str.split("\n")
        return script_list

    def _dedent(self, s: str) -> str:
        split = list(s.splitlines())
        prefix = os.path.commonprefix([line for line in split if line])
        indent_matches = re.findall(r"^\s+", prefix)
        if not indent_matches:
            return s
        indent_end_index = len(indent_matches[0])
        dedented_strings = [
            line[indent_end_index:] if line and line.startswith(prefix) else "" for line in split
        ]
        return "\n".join(dedented_strings)



import asyncio
import tempfile
from dataclasses import dataclass
from subprocess import CalledProcessError

from ofrak import Unpacker, Resource
from ofrak.core import (
    GenericBinary,
    FilesystemRoot,
    File,
    Folder,
    SpecialFileType,
    MagicDescriptionIdentifier,
)
from ofrak.model.component_model import ComponentExternalTool, ComponentConfig

_DEBUGFS = ComponentExternalTool(
    "debugfs", "https://e2fsprogs.sourceforge.net/", "-V", brew_package="e2fsprogs"
)


@dataclass
class ExtFilesystem(GenericBinary, FilesystemRoot):
    pass


@dataclass
class Ext2Filesystem(ExtFilesystem):
    """
    Linux EXT2 filesystem.
    """


@dataclass
class Ext3Filesystem(ExtFilesystem):
    """
    Linux EXT3 filesystem.
    """


@dataclass
class Ext4Filesystem(ExtFilesystem):
    """
    Linux EXT4 filesystem.
    """


class ExtUnpacker(Unpacker[None]):
    """
    Unpack a Linux EXT filesystem.
    """

    targets = (ExtFilesystem,)
    children = (File, Folder, SpecialFileType)
    external_dependencies = (_DEBUGFS,)

    async def unpack(self, resource: Resource, config: ComponentConfig = None) -> None:
        with tempfile.NamedTemporaryFile(suffix=".extfs") as temp_fs_file:
            temp_fs_file.write(await resource.get_data())
            temp_fs_file.flush()

            with tempfile.TemporaryDirectory() as temp_dir:
                command = [
                    "debugfs",
                    "-R",
                    f"rdump / {temp_dir}",
                    temp_fs_file.name,
                ]
                proc = await asyncio.create_subprocess_exec(
                    *command,
                )
                returncode = await proc.wait()
                if returncode:
                    raise CalledProcessError(returncode=returncode, cmd=command)

                fs_view = await resource.view_as(ExtFilesystem)
                await fs_view.initialize_from_disk(temp_dir)


MagicDescriptionIdentifier.register(Ext2Filesystem, lambda s: "ext2 filesystem" in s.lower())
MagicDescriptionIdentifier.register(Ext3Filesystem, lambda s: "ext3 filesystem" in s.lower())
MagicDescriptionIdentifier.register(Ext4Filesystem, lambda s: "ext4 filesystem" in s.lower())



from abc import ABC, abstractmethod


class EccError(Exception):
    pass


class EccAlgorithm(ABC):
    @abstractmethod
    def encode(self, payload: bytes) -> bytes:
        raise NotImplementedError()

    @abstractmethod
    def decode(self, payload: bytes) -> bytes:
        raise NotImplementedError()



import bisect
import itertools
import logging
import math
import sys
from abc import ABC, abstractmethod
from collections import defaultdict
from typing import Dict, List, Set, Optional, Iterable, Tuple, Any, TypeVar, Generic

from sortedcontainers import SortedList

from ofrak.model.resource_model import (
    ResourceModel,
    ResourceModelDiff,
    ResourceIndexedAttribute,
)
from ofrak.model.tag_model import ResourceTag
from ofrak.service.resource_service_i import (
    ResourceServiceInterface,
    ResourceFilter,
    ResourceSort,
    ResourceAttributeFilter,
    ResourceAttributeRangeFilter,
    ResourceAttributeValueFilter,
    ResourceAttributeValuesFilter,
    ResourceSortDirection,
    ResourceFilterCondition,
    ResourceServiceWalkError,
)
from ofrak_type.error import NotFoundError, AlreadyExistError
from ofrak_type.range import Range

LOGGER = logging.getLogger(__name__)
T = TypeVar("T", str, int, float, bytes)


class LowValue:
    def __gt__(self, other):
        return False


class HighValue:
    def __lt__(self, other):
        return False


LOW_VALUE = LowValue()
HIGH_VALUE = HighValue()


class ResourceNode:
    model: ResourceModel
    parent: Optional["ResourceNode"]
    _children: Dict["ResourceNode", None]
    _ancestor_ids: Dict[bytes, int]
    _descendant_count: int
    _depth: int

    def __init__(self, model: ResourceModel, parent: Optional["ResourceNode"]):
        self.model = model
        self.parent = parent
        # Dict serves as an ordered set to preserve children insertion order
        self._children: Dict[ResourceNode, None] = dict()
        self._ancestor_ids: Dict[bytes, int] = dict()
        self._descendant_count = 0
        self._depth = 0
        if self.parent is not None:
            self.parent.add_child(self)
            self.model.parent_id = self.parent.model.id
        else:
            self.model.parent_id = None

    def add_child(self, child: "ResourceNode"):
        child._depth = self._depth + 1

        child._ancestor_ids = {
            parent_id: parent_depth + 1 for parent_id, parent_depth in self._ancestor_ids.items()
        }
        child._ancestor_ids[self.model.id] = 1

        parent: Optional[ResourceNode] = self
        while parent is not None:
            parent._descendant_count += child._descendant_count + 1
            parent = parent.parent

        self._children[child] = None

    def remove_child(self, child: "ResourceNode"):
        del self._children[child]
        parent: Optional[ResourceNode] = self
        while parent is not None:
            parent._descendant_count -= child._descendant_count + 1
            parent = parent.parent

        ids_to_clear = list(child._ancestor_ids.keys())

        def remove_ancestor_ids(descendent: "ResourceNode"):
            for ancestor_id in ids_to_clear:
                del descendent._ancestor_ids[ancestor_id]

            for _descendent in descendent._children.keys():
                remove_ancestor_ids(_descendent)

        remove_ancestor_ids(child)

    def has_ancestor(self, id: bytes, max_depth: int = -1, include_self: bool = False) -> bool:
        if include_self and id == self.model.id:
            return True
        ancestor_depth = self._ancestor_ids.get(id)
        if ancestor_depth is None:
            return False
        if max_depth < 0:
            return True
        return ancestor_depth <= max_depth

    def walk_ancestors(self, include_self: bool) -> Iterable["ResourceNode"]:
        if include_self:
            yield self
        parent = self.parent
        while parent is not None:
            yield parent
            parent = parent.parent

    def get_depth(self) -> int:
        return self._depth

    def get_descendant_count(self) -> int:
        return self._descendant_count

    def walk_descendants(
        self, include_self: bool, max_depth: int, _depth: int = 0
    ) -> Iterable["ResourceNode"]:
        if include_self:
            yield self
        if 0 <= max_depth <= _depth:
            return
        for child in self._children.keys():
            yield from child.walk_descendants(True, max_depth, _depth + 1)

    def __lt__(self, other):
        if not isinstance(other, ResourceNode):
            return NotImplemented
        return self.model.id < other.model.id

    def __eq__(self, other):
        if not isinstance(other, ResourceNode):
            return False
        return self.model.id == other.model.id

    def __hash__(self):
        return hash(self.model.id)


class ResourceAttributeIndex(Generic[T]):
    def __init__(self, attribute: ResourceIndexedAttribute[T]):
        self._attribute: ResourceIndexedAttribute[T] = attribute
        self.index: SortedList = SortedList()
        self.values_by_node_id: Dict[bytes, Any] = dict()

    def add_resource_attribute(
        self,
        value: T,
        resource: ResourceNode,
    ):
        if resource.model.id in self.values_by_node_id:
            if self.values_by_node_id[resource.model.id] != value:
                raise ValueError(
                    f"The provided resource {resource.model.id.hex()} is already in the "
                    f"index for {self._attribute.__name__} with a different value!"
                )
            else:
                return
        self.index.add((value, resource))
        self.values_by_node_id[resource.model.id] = value

    def remove_resource_attribute(
        self,
        resource: ResourceNode,
    ):
        if resource.model.id not in self.values_by_node_id:
            return
        value = self.values_by_node_id[resource.model.id]
        self.index.remove((value, resource))
        del self.values_by_node_id[resource.model.id]


class AttributeIndexDict(defaultdict):
    """
    `defaultdict` that passes the missing key to the default factory.

    See:
    <https://stackoverflow.com/questions/2912231/is-there-a-clever-way-to-pass-the-key-to-defaultdicts-default-factory>
    """

    def __missing__(self, key):
        if self.default_factory is None:
            raise KeyError(key)
        else:
            ret = self[key] = self.default_factory(key)
            return ret


class ResourceSortLogic(Generic[T], ABC):
    @abstractmethod
    def has_effect(self) -> bool:
        raise NotImplementedError()

    @abstractmethod
    def get_match_count(self) -> int:
        raise NotImplementedError()

    @abstractmethod
    def _get_attribute_value(self, resource: ResourceModel) -> Optional[T]:
        raise NotImplementedError()

    @abstractmethod
    def sort(self, resources: Iterable[ResourceModel]) -> Iterable[ResourceModel]:
        raise NotImplementedError()

    @abstractmethod
    def walk(self) -> Iterable[ResourceNode]:
        raise NotImplementedError()

    @abstractmethod
    def get_attribute(self) -> Optional[ResourceIndexedAttribute[T]]:
        raise NotImplementedError()

    @abstractmethod
    def get_direction(self) -> ResourceSortDirection:
        raise NotImplementedError()

    @staticmethod
    def create(
        r_sort: Optional[ResourceSort],
        attribute_indexes: Dict[ResourceIndexedAttribute[T], ResourceAttributeIndex],
    ) -> "ResourceSortLogic[T]":
        if r_sort is None:
            return NullResourceSortLogic()
        attribute_index = attribute_indexes[r_sort.attribute].index
        return ActiveResourceSortLogic[T](r_sort.attribute, attribute_index, r_sort.direction)


class ActiveResourceSortLogic(ResourceSortLogic[T]):
    def __init__(
        self,
        attribute: ResourceIndexedAttribute[T],
        index: List[Tuple[Any, ResourceNode]],
        direction: ResourceSortDirection = ResourceSortDirection.ASCENDANT,
    ):
        self.attribute: ResourceIndexedAttribute[T] = attribute  # type: ignore
        self.index: List[Tuple[Any, ResourceNode]] = index
        self.direction: ResourceSortDirection = direction

    def has_effect(self) -> bool:
        return True

    def get_match_count(self) -> int:
        return len(self.index)

    def get_attribute(self) -> ResourceIndexedAttribute[T]:
        return self.attribute

    def get_direction(self) -> ResourceSortDirection:
        return self.direction

    def _get_attribute_value(self, resource: ResourceModel) -> T:
        value = self.attribute.get_value(resource)
        if value is None:
            raise ValueError()
        else:
            return value

    def sort(self, resources: Iterable[ResourceModel]) -> Iterable[ResourceModel]:
        if self.attribute is None:
            raise ValueError("No attribute specified to sort on")
        reverse = self.direction != ResourceSortDirection.ASCENDANT
        return sorted(resources, key=self._get_attribute_value, reverse=reverse)

    def walk(self) -> Iterable[ResourceNode]:
        if self.index is None:
            raise ResourceServiceWalkError("Cannot walk a ResourceSortLogic with no index!")
        if self.direction is ResourceSortDirection.ASCENDANT:
            index = 0
            increment = 1
        else:
            index = len(self.index) - 1
            increment = -1
        max_index = len(self.index)
        while 0 <= index < max_index:
            yield self.index[index][1]
            index += increment


class NullResourceSortLogic(ResourceSortLogic):
    def has_effect(self) -> bool:
        return False

    def get_attribute(self) -> None:
        return None

    def get_match_count(self) -> int:  # pragma: no cover
        raise NotImplementedError()

    def _get_attribute_value(self, resource: ResourceModel) -> None:  # pragma: no cover
        raise NotImplementedError()

    def sort(
        self, resources: Iterable[ResourceModel]
    ) -> Iterable[ResourceModel]:  # pragma: no cover
        raise NotImplementedError()

    def walk(self) -> Iterable[ResourceNode]:  # pragma: no cover
        raise NotImplementedError()

    def get_direction(self) -> ResourceSortDirection:  # pragma: no cover
        raise NotImplementedError()  # pragma: no cover


class ResourceFilterLogic(Generic[T], ABC):
    def get_attribute(self) -> Optional[ResourceIndexedAttribute[T]]:
        return None

    @abstractmethod
    def filter(self, value: ResourceNode) -> bool:
        pass

    @abstractmethod
    def get_match_count(self) -> int:
        pass

    @abstractmethod
    def walk(self, direction: ResourceSortDirection) -> Iterable[ResourceNode]:
        pass

    @classmethod
    def get_attribute_value(
        cls, resource: ResourceNode, attribute_type: ResourceIndexedAttribute[T]
    ) -> Optional[T]:
        return attribute_type.get_value(resource.model)


class ResourceAttributeFilterLogic(ResourceFilterLogic[T], ABC):
    def __init__(
        self,
        attribute: ResourceIndexedAttribute[T],
        index: List[Tuple[T, ResourceNode]],
    ):
        self.attribute: ResourceIndexedAttribute[T] = attribute  # type: ignore
        self.index: List[Tuple[T, ResourceNode]] = index
        self._cached_ranges: Optional[Tuple[Range, ...]] = None

    def get_attribute(self) -> Optional[ResourceIndexedAttribute[T]]:
        return self.attribute

    @abstractmethod
    def _compute_ranges(self) -> Iterable[Range]:
        raise NotImplementedError()

    def walk(self, direction: ResourceSortDirection) -> Iterable[ResourceNode]:
        if self._cached_ranges is None:
            self._cached_ranges = tuple(sorted(self._compute_ranges(), key=lambda r: r.start))
        cached_ranges: Iterable[Range] = ()
        if direction is ResourceSortDirection.ASCENDANT:
            cached_ranges = self._cached_ranges
        else:
            cached_ranges = tuple(reversed(self._cached_ranges))

        for index_range in cached_ranges:
            if direction is ResourceSortDirection.ASCENDANT:
                index = index_range.start
                increment = 1
            else:
                index = index_range.end - 1
                increment = -1
            min_index = index_range.start
            max_index = index_range.end
            while min_index <= index < max_index:
                yield self.index[index][1]
                index += increment

    def get_match_count(self) -> int:
        if self._cached_ranges is None:
            self._cached_ranges = tuple(sorted(self._compute_ranges(), key=lambda r: r.start))
        return sum(r.length() for r in self._cached_ranges)


class ResourceAttributeRangeFilterLogic(ResourceAttributeFilterLogic, Generic[T]):
    def __init__(
        self,
        attribute: ResourceIndexedAttribute[T],
        index: List[Tuple[T, ResourceNode]],
        min: T = None,
        max: T = None,
    ):
        if min is None and max is None:
            raise ValueError("Invalid filter, either a min, a max or both must be provided")
        super().__init__(attribute, index)
        self.min: Optional[T] = min
        self.max: Optional[T] = max

    def _compute_ranges(self) -> Iterable[Range]:
        if self.min is not None:
            min_index = bisect.bisect_left(self.index, (self.min, LOW_VALUE))
        else:
            min_index = 0
        if self.max is not None:
            # TODO: There should most likely be a +1 in here
            max_index = bisect.bisect_left(self.index, (self.max, LOW_VALUE))
        else:
            max_index = len(self.index)
        return (Range(min_index, max_index),)

    def filter(self, resource: ResourceNode) -> bool:
        value = self.get_attribute_value(resource, self.attribute)
        if value is None:
            return False
        if self.min is not None and self.max is not None:
            return self.min <= value < self.max
        if self.min is not None:
            return self.min <= value
        elif self.max is not None:
            return value < self.max
        else:
            raise ValueError("Invalid filter, either a min, a max or both must be provided")


class ResourceAttributeValueFilterLogic(ResourceAttributeFilterLogic, Generic[T]):
    def __init__(
        self,
        attribute: ResourceIndexedAttribute[T],
        index: List[Tuple[T, ResourceNode]],
        value: T,
    ):
        super().__init__(attribute, index)
        self.value: T = value

    def _compute_ranges(self) -> Iterable[Range]:
        return (
            Range(
                bisect.bisect_left(self.index, (self.value, LOW_VALUE)),
                bisect.bisect_right(self.index, (self.value, HIGH_VALUE)),
            ),
        )

    def filter(self, resource: ResourceNode) -> bool:
        value = self.get_attribute_value(resource, self.attribute)
        if value is None:
            return False
        return value == self.value


class ResourceAttributeValuesFilterLogic(ResourceAttributeFilterLogic, Generic[T]):
    def __init__(
        self,
        attribute: ResourceIndexedAttribute[T],
        index: List[Tuple[T, ResourceNode]],
        values: Tuple[T, ...],
    ):
        super().__init__(attribute, index)
        self.index = index
        self.values: Set[T] = set(values)

    def _compute_ranges(self) -> Iterable[Range]:
        for value in self.values:
            yield Range(
                bisect.bisect_left(self.index, (value, LOW_VALUE)),
                bisect.bisect_right(self.index, (value, HIGH_VALUE)),
            )

    def filter(self, resource: ResourceNode) -> bool:
        value = self.get_attribute_value(resource, self.attribute)
        if value is None:
            return False
        return value in self.values


class ResourceTagOrFilterLogic(ResourceFilterLogic):
    def __init__(
        self,
        indexes: Dict[ResourceTag, Set[ResourceNode]],
        tags: Tuple[ResourceTag, ...],
    ):
        if len(tags) == 0:
            raise ValueError(
                "Cannot instantiate the ResourceTagOrFilterLogic class with an empty set of tags "
                "to filter on."
            )
        self.indexes = indexes
        self.tags = tags

    def filter(self, resource: ResourceNode) -> bool:
        for tag in self.tags:
            if resource.model.has_tag(tag):
                return True
        return False

    def get_match_count(self) -> int:
        count = 0
        for tag in self.tags:
            count += len(self.indexes[tag])
        return count

    def walk(self, direction: ResourceSortDirection) -> Iterable[ResourceNode]:
        processed_ids = set()
        for tag in self.tags:
            for resource in self.indexes[tag]:
                resource_m = resource.model
                if resource_m.id in processed_ids:
                    continue
                processed_ids.add(resource_m.id)
                yield resource


class ResourceTagAndFilterLogic(ResourceFilterLogic):
    def __init__(
        self,
        indexes: Dict[ResourceTag, Set[ResourceNode]],
        tags: Tuple[ResourceTag, ...],
    ):
        if len(tags) == 0:
            raise ValueError(
                "Cannot instantiate the ResourceTagAndFilterLogic class with an empty set of tags "
                "to filter on."
            )
        self.indexes: Dict[ResourceTag, Set[ResourceNode]] = indexes
        self.tags = tags

        self._walk_tag: Optional[ResourceTag] = None
        self._filter_tags: Optional[Tuple[ResourceTag, ...]] = None

    def filter(self, resource: ResourceNode) -> bool:
        for tag in self.tags:
            if not resource.model.has_tag(tag):
                return False
        return True

    def _compute_tags(self) -> Tuple[ResourceTag, Optional[Tuple[ResourceTag, ...]]]:
        if self._walk_tag is not None:
            return self._walk_tag, self._filter_tags
        min_size = sys.maxsize
        walk_tag = None
        for tag in self.tags:
            index_size = len(self.indexes[tag])
            if index_size < min_size:
                walk_tag = tag
                min_size = index_size
        if walk_tag is None:
            # No tags in self.tags had fewer than sys.maxsize resources with that tag. Choose one
            # arbitrarily to be the walk tag then, since they all have equal (very high) cost.
            walk_tag = self.tags[0]
        self._walk_tag = walk_tag
        filter_tags = tuple(filter(lambda t: t != walk_tag, self.tags))
        self._filter_tags = filter_tags

        return walk_tag, filter_tags

    def get_match_count(self) -> int:
        walk_tag, _ = self._compute_tags()
        return len(self.indexes[walk_tag])

    def walk(self, direction: ResourceSortDirection) -> Iterable[ResourceNode]:
        walk_tag, filter_tags = self._compute_tags()
        main_index = self.indexes[walk_tag]
        for resource in main_index:
            if filter_tags is not None and len(filter_tags) > 0:
                for tag in filter_tags:
                    if resource in self.indexes[tag]:
                        yield resource
            else:
                yield resource


class ResourceAncestorFilterLogic(ResourceFilterLogic):
    def __init__(
        self,
        root: ResourceNode,
        include_root: bool = False,
        max_depth: int = -1,
    ):
        self.root = root
        self.include_root = include_root
        self.max_depth = max_depth

    def filter(self, resource: ResourceNode) -> bool:
        return resource.has_ancestor(self.root.model.id, self.max_depth, self.include_root)

    def get_match_count(self) -> int:
        count = self.root.get_descendant_count()
        if self.include_root:
            count += 1
        return count

    def walk(self, direction: ResourceSortDirection) -> Iterable[ResourceNode]:
        return self.root.walk_descendants(include_self=self.include_root, max_depth=self.max_depth)


class AggregateResourceFilterLogic:
    def __init__(self, filters: Tuple[ResourceFilterLogic, ...]):
        self.filters = filters

    @staticmethod
    def _create_attribute_filter(
        attribute_filter: ResourceAttributeFilter, attribute_index: List[Tuple[T, ResourceNode]]
    ) -> ResourceFilterLogic[T]:
        if isinstance(attribute_filter, ResourceAttributeRangeFilter):
            return ResourceAttributeRangeFilterLogic(
                attribute_filter.attribute,
                attribute_index,
                attribute_filter.min,
                attribute_filter.max,
            )
        elif isinstance(attribute_filter, ResourceAttributeValueFilter):
            return ResourceAttributeValueFilterLogic(
                attribute_filter.attribute, attribute_index, attribute_filter.value
            )
        elif isinstance(attribute_filter, ResourceAttributeValuesFilter):
            return ResourceAttributeValuesFilterLogic(
                attribute_filter.attribute, attribute_index, attribute_filter.values
            )
        else:
            raise ValueError(f"Unknown filter of type {type(attribute_filter).__name__}")

    @staticmethod
    def create(
        r_filter: Optional[ResourceFilter],
        tag_indexes: Dict[ResourceTag, Set[ResourceNode]],
        attribute_indexes: Dict[ResourceIndexedAttribute[T], ResourceAttributeIndex[T]],
        ancestor: ResourceNode = None,
        max_depth: int = -1,
    ) -> "AggregateResourceFilterLogic":
        filters: List[ResourceFilterLogic] = []
        if r_filter is not None:
            if r_filter.tags is not None:
                if r_filter.tags_condition is ResourceFilterCondition.AND:
                    filters.append(ResourceTagAndFilterLogic(tag_indexes, tuple(r_filter.tags)))
                else:
                    filters.append(ResourceTagOrFilterLogic(tag_indexes, tuple(r_filter.tags)))
            if r_filter.attribute_filters is not None:
                for attribute_filter in r_filter.attribute_filters:
                    filters.append(
                        AggregateResourceFilterLogic._create_attribute_filter(
                            attribute_filter, attribute_indexes[attribute_filter.attribute].index
                        )
                    )

            include_root = r_filter.include_self
        else:
            include_root = False
        if ancestor is not None:
            filters.append(ResourceAncestorFilterLogic(ancestor, include_root, max_depth))
        return AggregateResourceFilterLogic(tuple(filters))

    def ignore_filter(self, filter_logic: ResourceFilterLogic):
        if filter_logic is None:
            raise ValueError("Invalid index filter logic")
        self.filters = tuple(ix for ix in self.filters if ix != filter_logic)

    def has_effect(self) -> bool:
        return len(self.filters) > 0

    def filter(self, resource: ResourceNode) -> bool:
        for filter_logic in self.filters:
            if not filter_logic.filter(resource):
                return False
        return True


class ResourceService(ResourceServiceInterface):
    def __init__(self):
        self._resource_store: Dict[bytes, ResourceNode] = dict()
        self._resource_by_data_id_store: Dict[bytes, ResourceNode] = dict()
        self._attribute_indexes: Dict[
            ResourceIndexedAttribute[T], ResourceAttributeIndex[T]
        ] = AttributeIndexDict(ResourceAttributeIndex)
        self._tag_indexes: Dict[ResourceTag, Set[ResourceNode]] = defaultdict(set)
        self._root_resources: Dict[bytes, ResourceNode] = dict()

    def _add_resource_tag_to_index(self, tag: ResourceTag, resource: ResourceNode):
        for _tag in tag.tag_classes():
            self._tag_indexes[_tag].add(resource)

    def _remove_resource_tag_from_index(
        self,
        tag: ResourceTag,
        resource: ResourceNode,
        blacklist: Set[ResourceTag],
    ):
        for _tag in tag.tag_classes():
            if blacklist is not None and _tag in blacklist:
                continue
            self._tag_indexes[_tag].remove(resource)

    def _add_resource_attribute_to_index(
        self,
        indexable_attribute: ResourceIndexedAttribute[T],
        value: T,
        resource: ResourceNode,
    ):
        if value is None:
            return
        index = self._attribute_indexes[indexable_attribute]
        index.add_resource_attribute(value, resource)

        for dependent_indexable in indexable_attribute.used_by_indexes:
            dependant_value = dependent_indexable.get_value(resource.model)
            self._add_resource_attribute_to_index(dependent_indexable, dependant_value, resource)

    def _remove_resource_attribute_from_index(
        self,
        indexable_attribute: ResourceIndexedAttribute[T],
        resource: ResourceNode,
    ):
        index = self._attribute_indexes[indexable_attribute]
        index.remove_resource_attribute(resource)
        for dependent_indexable in indexable_attribute.used_by_indexes:
            self._remove_resource_attribute_from_index(dependent_indexable, resource)

    async def create(self, resource: ResourceModel) -> ResourceModel:
        if resource.id in self._resource_store:
            raise AlreadyExistError(f"A resource with id {resource.id.hex()} already exists!")
        if resource.parent_id is not None:
            parent_resource_node = self._resource_store.get(resource.parent_id)
            if parent_resource_node is None:
                raise NotFoundError(
                    f"The parent resource with id {resource.parent_id.hex()} does not exist"
                )
            LOGGER.debug(
                f"Creating resource {resource.id.hex()} as child of {resource.parent_id.hex()}"
            )
        else:
            parent_resource_node = None
            LOGGER.debug(f"Creating resource {resource.id.hex()}")
        resource_node = ResourceNode(resource, parent_resource_node)
        self._resource_store[resource.id] = resource_node
        if resource.data_id is not None:
            self._resource_by_data_id_store[resource.data_id] = resource_node
        if parent_resource_node is None:
            self._root_resources[resource.id] = resource_node

        # Take care of the indexes
        for tag in resource.tags:
            self._add_resource_tag_to_index(tag, resource_node)
        for indexable_attribute, value in resource.get_index_values().items():
            self._add_resource_attribute_to_index(indexable_attribute, value, resource_node)
        return resource

    async def get_root_resources(self) -> List[ResourceModel]:
        return [root_node.model for root_node in self._root_resources.values()]

    async def verify_ids_exist(self, resource_ids: Iterable[bytes]) -> Iterable[bool]:
        return [resource_id in self._resource_store for resource_id in resource_ids]

    async def get_by_data_ids(self, data_ids: Iterable[bytes]) -> Iterable[ResourceModel]:
        results = []
        for data_id in data_ids:
            resource_node = self._resource_by_data_id_store.get(data_id)
            if resource_node is None:
                raise NotFoundError(f"The resource with data ID {data_id.hex()} does not exist")
            results.append(resource_node.model)
        return results

    async def get_by_ids(self, resource_ids: Iterable[bytes]) -> Iterable[ResourceModel]:
        results = []
        for resource_id in resource_ids:
            resource_node = self._resource_store.get(resource_id)
            if resource_node is None:
                raise NotFoundError(f"The resource {resource_id.hex()} does not exist")
            results.append(resource_node.model)
        return results

    async def get_by_id(self, resource_id: bytes) -> ResourceModel:
        LOGGER.debug(f"Fetching resource {resource_id.hex()}")
        resource_node = self._resource_store.get(resource_id)
        if resource_node is None:
            raise NotFoundError(f"The resource {resource_id.hex()} does not exist")
        return resource_node.model

    async def get_depths(self, resource_ids: Iterable[bytes]) -> Iterable[int]:
        results = []
        for resource_id in resource_ids:
            resource_node = self._resource_store.get(resource_id)
            if resource_node is None:
                raise NotFoundError(f"The resource {resource_id.hex()} does not exist")
            results.append(resource_node.get_depth())
        return results

    async def get_ancestors_by_id(
        self,
        resource_id: bytes,
        max_count: int = -1,
        r_filter: Optional[ResourceFilter] = None,
    ) -> Iterable[ResourceModel]:
        LOGGER.debug(f"Fetching ancestor(s) of {resource_id.hex()}")
        resource_node = self._resource_store.get(resource_id)
        if resource_node is None:
            raise NotFoundError(f"The resource {resource_id.hex()} does not exist")
        r_filter_logic = AggregateResourceFilterLogic.create(
            r_filter,
            self._tag_indexes,
            self._attribute_indexes,
        )
        include_root = False if r_filter is None else r_filter.include_self
        resources = map(
            lambda n: n.model,
            filter(r_filter_logic.filter, resource_node.walk_ancestors(include_root)),
        )
        if max_count < 0:
            return resources
        return itertools.islice(resources, 0, max_count)

    async def get_descendants_by_id(
        self,
        resource_id: bytes,
        max_count: int = -1,
        max_depth: int = -1,
        r_filter: Optional[ResourceFilter] = None,
        r_sort: Optional[ResourceSort] = None,
    ) -> Iterable[ResourceModel]:
        # LOGGER.debug(f"Fetching descendant(s) of {resource_id.hex()}")
        resource_node = self._resource_store.get(resource_id)
        if resource_node is None:
            raise NotFoundError(f"The resource {resource_id.hex()} does not exist")

        aggregate_sort_logic = ResourceSortLogic.create(r_sort, self._attribute_indexes)
        aggregate_filter_logic = AggregateResourceFilterLogic.create(
            r_filter, self._tag_indexes, self._attribute_indexes, resource_node, max_depth
        )
        # This is the planning phase used to determine the best index to use for further filtering
        filter_logic: Optional[ResourceFilterLogic] = None
        filter_cost = sys.maxsize
        sort_cost = sys.maxsize
        if aggregate_sort_logic.has_effect():
            sort_cost = aggregate_sort_logic.get_match_count()
            if sort_cost == 0:
                return tuple()

        for _filter_logic in aggregate_filter_logic.filters:
            _filter_cost = _filter_logic.get_match_count()
            if _filter_cost == 0:
                return tuple()
            if (
                aggregate_sort_logic.has_effect()
                and aggregate_sort_logic.get_attribute() != _filter_logic.get_attribute()
            ):
                # The resources matching the filter would need to get sorted, making the
                # worst case # scenario more expensive
                _filter_cost = int(_filter_cost * math.log2(_filter_cost))
            if _filter_cost < filter_cost:
                filter_cost = _filter_cost
                filter_logic = _filter_logic

        # Use the estimated cost to pick the fastest way to compute the results
        if (
            filter_logic is not None
            and filter_logic.get_attribute() is not None
            and filter_logic.get_attribute() == aggregate_sort_logic.get_attribute()
        ):
            resource_nodes = filter_logic.walk(aggregate_sort_logic.get_direction())
            aggregate_filter_logic.ignore_filter(filter_logic)
            aggregate_sort_logic = NullResourceSortLogic()
        elif sort_cost < filter_cost:
            resource_nodes = aggregate_sort_logic.walk()
            aggregate_sort_logic = NullResourceSortLogic()
        elif filter_logic is not None:
            resource_nodes = filter_logic.walk(ResourceSortDirection.ASCENDANT)
            # No need to filter on that index since it serves as the root index
            aggregate_filter_logic.ignore_filter(filter_logic)

        if aggregate_filter_logic.has_effect():
            resource_nodes = filter(aggregate_filter_logic.filter, resource_nodes)
        resources: Iterable[ResourceModel] = map(lambda n: n.model, resource_nodes)
        if aggregate_sort_logic.has_effect():
            resources = aggregate_sort_logic.sort(resources)
        if max_count >= 0:
            resources = itertools.islice(resources, 0, max_count)
        return resources

    async def get_siblings_by_id(
        self,
        resource_id: bytes,
        max_count: int = -1,
        r_filter: Optional[ResourceFilter] = None,
        r_sort: Optional[ResourceSort] = None,
    ) -> Iterable[ResourceModel]:
        resource_node = self._resource_store.get(resource_id)
        if resource_node is None:
            raise NotFoundError(f"The resource {resource_id.hex()} does not exist")
        if resource_node.parent is None:
            raise NotFoundError(
                f"The resource {resource_id.hex()} does not have siblings as it is a root "
                f"resource."
            )
        return await self.get_descendants_by_id(
            resource_node.parent.model.id, max_count, 1, r_filter, r_sort
        )

    async def update(self, resource_diff: ResourceModelDiff) -> ResourceModel:
        return self._update(resource_diff)

    async def update_many(
        self, resource_diffs: Iterable[ResourceModelDiff]
    ) -> Iterable[ResourceModel]:
        return [self._update(resource_diff) for resource_diff in resource_diffs]

    def _update(self, resource_diff: ResourceModelDiff) -> ResourceModel:
        LOGGER.debug(f"Saving resource {resource_diff.id.hex()}")
        resource_node = self._resource_store.get(resource_diff.id)
        if resource_node is None:
            raise NotFoundError(f"The resource with ID {resource_diff.id.hex()} does not exist")

        prev_resource = resource_node.model
        next_resource = resource_diff.apply(prev_resource)

        current_tags = next_resource.get_tags()
        # Update the tag indexes
        for tag_removed in resource_diff.tags_removed:
            self._remove_resource_tag_from_index(tag_removed, resource_node, set(current_tags))
        for tag_added in resource_diff.tags_added:
            self._add_resource_tag_to_index(tag_added, resource_node)

        # Update the attribute indexes
        indexable_attributes_removed = set()
        for attributes_removed in resource_diff.attributes_removed:
            for indexable_attribute_removed in attributes_removed.get_indexable_attributes():
                indexable_attributes_removed.add(indexable_attribute_removed)
                self._remove_resource_attribute_from_index(
                    indexable_attribute_removed, resource_node
                )

        indexable_attributes_added = set()
        for attributes_type_added, attributes_added in resource_diff.attributes_added.items():
            for indexable_attribute_added in attributes_added.get_indexable_attributes():
                indexable_attributes_added.add(indexable_attribute_added)
                self._add_resource_attribute_to_index(
                    indexable_attribute_added,
                    indexable_attribute_added.get_value(next_resource),
                    resource_node,
                )

        resource_node.model = next_resource
        return next_resource

    async def rebase_resource(self, resource_id: bytes, new_parent_id: bytes):
        resource_node = self._resource_store.get(resource_id)
        if resource_node is None:
            raise NotFoundError(f"The resource {resource_id.hex()} does not exist")

        new_parent_resource_node = self._resource_store.get(new_parent_id)
        if new_parent_resource_node is None:
            raise NotFoundError(f"The new parent resource {resource_id.hex()} does not exist")

        former_parent_resource_node = resource_node.parent
        if former_parent_resource_node is not None:
            former_parent_resource_node.remove_child(resource_node)
        new_parent_resource_node.add_child(resource_node)
        resource_node.parent = new_parent_resource_node
        resource_node.model.parent_id = new_parent_id

    async def delete_resource(self, resource_id: bytes):
        resource_node = self._resource_store.get(resource_id)
        if resource_node is None:
            # Already deleted, probably by an ancestor calling the recursive func below
            return []

        former_parent_resource_node = resource_node.parent
        if former_parent_resource_node is not None:
            former_parent_resource_node.remove_child(resource_node)

        deleted_models = self._delete_resource_helper(resource_node)
        LOGGER.debug(f"Deleted {resource_id.hex()}")
        return deleted_models

    async def delete_resources(self, resource_ids: Iterable[bytes]):
        deleted_models = []
        for resource_id in resource_ids:
            resource_node = self._resource_store.get(resource_id)
            if resource_node is None:
                # Already deleted, probably by an ancestor calling the recursive func below
                continue

            former_parent_resource_node = resource_node.parent
            if former_parent_resource_node is not None:
                former_parent_resource_node.remove_child(resource_node)

            deleted_models.extend(self._delete_resource_helper(resource_node))
        if resource_ids:
            LOGGER.debug(f"Deleted {', '.join(resource_id.hex() for resource_id in resource_ids)}")
        return deleted_models

    def _delete_resource_helper(self, _resource_node: ResourceNode):
        deleted_models = []
        for child in _resource_node._children.keys():
            deleted_models.extend(self._delete_resource_helper(child))

        for indexable_attribute, val in _resource_node.model.get_index_values().items():
            self._remove_resource_attribute_from_index(indexable_attribute, _resource_node)

        tag_removal_blacklist: Set[ResourceTag] = set()
        for tag in _resource_node.model.tags:
            self._remove_resource_tag_from_index(tag, _resource_node, tag_removal_blacklist)
            tag_removal_blacklist.update(tag.tag_classes())

        del self._resource_store[_resource_node.model.id]
        if _resource_node.model.data_id is not None:
            del self._resource_by_data_id_store[_resource_node.model.data_id]

        deleted_models.append(_resource_node.model)
        return deleted_models



from dataclasses import dataclass, field
from typing import Any, Dict, Union

from ofrak.component.modifier import Modifier
from ofrak.model.component_model import ComponentConfig
from ofrak.resource import Resource

_PrimitiveType = Union[int, bool, str, float, bytes]


@dataclass
class RunScriptModifierConfig(ComponentConfig):
    """
    Specification of some Python code to run in the current OFRAK context. Useful for reusing quick
    helper functions and the like.

    :ivar code: Python source code defining one or more function which take an OFRAKContext and
    Resource as arguments.
    :ivar function_name: Name of the function to run.
    :ivar extra_args: Extra arguments to pass to the function in key-value form.
    """

    code: str
    function_name: str = "main"
    extra_args: Dict[str, _PrimitiveType] = field(default_factory=dict)


class RunScriptModifier(Modifier[RunScriptModifierConfig]):
    """
    "Import" and run Python functions in the current OFRAK context. Useful for reusing quick
    helper functions and the like. Since this can be run through the GUI, it can be used to
    automate tasks which might be repetitive or impossible purely through the graphical interface.

    """

    targets = ()

    async def modify(self, resource: Resource, config: RunScriptModifierConfig) -> None:
        script_globals: Dict[str, Any] = dict()
        script_locals: Dict[str, Any] = dict()
        exec(config.code, script_globals, script_locals)

        if config.function_name in script_globals:
            script_main = script_globals[config.function_name]
        elif config.function_name in script_locals:
            script_main = script_locals[config.function_name]
        else:
            raise ValueError(f"No `{config.function_name}` function found in script!")

        from ofrak.ofrak_context import get_current_ofrak_context

        context = get_current_ofrak_context()

        script_main.__globals__.update(script_globals)
        script_main.__globals__.update(script_locals)

        full_kwargs: Dict[str, Any] = {
            "ofrak_context": context,
            "root_resource": resource,
        }
        full_kwargs.update(config.extra_args)

        await script_main(**full_kwargs)



import inspect
import logging


LOGGER = logging.getLogger(__name__)


class xattr:
    """
    Stub library to support OFRAK on Windows and other platforms where xattr is not available.
    """

    def __init__(self, obj, options=0):
        frame = inspect.currentframe()
        _warn_user_no_xattr(inspect.getframeinfo(frame).function)

    def __repr__(self):
        frame = inspect.currentframe()
        _warn_user_no_xattr(inspect.getframeinfo(frame).function)
        return ""

    def get(self, name, options=0):
        frame = inspect.currentframe()
        _warn_user_no_xattr(inspect.getframeinfo(frame).function)
        return b""

    def set(self, name, value, options=0):
        frame = inspect.currentframe()
        _warn_user_no_xattr(inspect.getframeinfo(frame).function)
        return None

    def remove(self, name, options=0):
        frame = inspect.currentframe()
        _warn_user_no_xattr(inspect.getframeinfo(frame).function)
        return None

    def list(self, options=0):
        frame = inspect.currentframe()
        _warn_user_no_xattr(inspect.getframeinfo(frame).function)
        return []

    def __len__(self):
        frame = inspect.currentframe()
        _warn_user_no_xattr(inspect.getframeinfo(frame).function)
        return 0

    def __delitem__(self, item):
        frame = inspect.currentframe()
        _warn_user_no_xattr(inspect.getframeinfo(frame).function)
        return None

    def __setitem__(self, item, value):
        frame = inspect.currentframe()
        _warn_user_no_xattr(inspect.getframeinfo(frame).function)
        return None

    def __getitem__(self, item):
        frame = inspect.currentframe()
        _warn_user_no_xattr(inspect.getframeinfo(frame).function)
        return b""

    def iterkeys(self):
        frame = inspect.currentframe()
        _warn_user_no_xattr(inspect.getframeinfo(frame).function)
        return iter(list())

    def has_key(self, item):
        frame = inspect.currentframe()
        _warn_user_no_xattr(inspect.getframeinfo(frame).function)
        return False

    def clear(self):
        frame = inspect.currentframe()
        _warn_user_no_xattr(inspect.getframeinfo(frame).function)
        return None

    def update(self, seq):
        frame = inspect.currentframe()
        _warn_user_no_xattr(inspect.getframeinfo(frame).function)
        return None

    def copy(self):
        frame = inspect.currentframe()
        _warn_user_no_xattr(inspect.getframeinfo(frame).function)
        return dict()

    def setdefault(self, k, d=""):
        frame = inspect.currentframe()
        _warn_user_no_xattr(inspect.getframeinfo(frame).function)
        return b""

    def keys(self):
        frame = inspect.currentframe()
        _warn_user_no_xattr(inspect.getframeinfo(frame).function)
        return []

    def itervalues(self):
        frame = inspect.currentframe()
        _warn_user_no_xattr(inspect.getframeinfo(frame).function)
        yield b""

    def values(self):
        frame = inspect.currentframe()
        _warn_user_no_xattr(inspect.getframeinfo(frame).function)
        return []

    def iteritems(self):
        frame = inspect.currentframe()
        _warn_user_no_xattr(inspect.getframeinfo(frame).function)
        yield tuple()

    def items(self):
        frame = inspect.currentframe()
        _warn_user_no_xattr(inspect.getframeinfo(frame).function)
        return []


def listxattr(f, symlink=False):
    frame = inspect.currentframe()
    _warn_user_no_xattr(inspect.getframeinfo(frame).function)
    return tuple()


def getxattr(f, attr, symlink=False):
    frame = inspect.currentframe()
    _warn_user_no_xattr(inspect.getframeinfo(frame).function)
    return b""


def setxattr(f, attr, value, options=0, symlink=False):
    frame = inspect.currentframe()
    _warn_user_no_xattr(inspect.getframeinfo(frame).function)
    return None


def removexattr(f, attr, symlink=False):
    frame = inspect.currentframe()
    _warn_user_no_xattr(inspect.getframeinfo(frame).function)
    return None


def _warn_user_no_xattr(function_name: str) -> None:
    LOGGER.warning(
        f"Function {function_name} not found. Library xattr is not available on Windows platforms. \
        Extended attributes will not be properly handled while using OFRAK on this platform. \
        If you require extended attributes, please use a platform that supports xattr."
    )



from abc import ABC, abstractmethod
from typing import Generic, Optional, Tuple

from ofrak.model.component_model import CC, ComponentRunResult, ComponentExternalTool
from ofrak.model.job_model import JobRunContext
from ofrak.model.resource_model import ResourceContext
from ofrak.model.tag_model import ResourceTag
from ofrak.model.viewable_tag_model import ResourceViewContext


class ComponentInterface(Generic[CC], ABC):
    """
    A component is responsible for modifying or create firmware resources. They are categorized
    as identifiers, unpackers, analyzers, modifiers and packers.
    """

    version: Optional[int] = None
    id: Optional[bytes] = None

    @classmethod
    @abstractmethod
    def get_default_config(cls) -> Optional[CC]:
        raise NotImplementedError()

    @abstractmethod
    def get_version(self) -> int:
        raise NotImplementedError()

    @property
    @abstractmethod
    def targets(self) -> Tuple[ResourceTag, ...]:
        raise NotImplementedError()

    @property
    @abstractmethod
    def external_dependencies(self) -> Tuple[ComponentExternalTool, ...]:
        raise NotImplementedError()

    @abstractmethod
    async def run(
        self,
        job_id: bytes,
        resource_id: bytes,
        job_context: JobRunContext,
        resource_context: ResourceContext,
        resource_view_context: ResourceViewContext,
        config: CC,
    ) -> ComponentRunResult:
        """

        :param job_id:
        :param resource_id:
        :param job_context:
        :param resource_context:
        :param resource_view_context:
        :param config:
        :return: The IDs of all resources modified by this component
        """
        raise NotImplementedError()

    @classmethod
    @abstractmethod
    def get_id(cls) -> bytes:
        raise NotImplementedError()
